/**
    PSPComic - pspcomic.c
    Copyright (C) 2007  Jeffrey P.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**/
#include "pspcomic.h"

#ifdef __cplusplus
extern "C" {
#endif

static int initial_clock = 222;
static char *cwd = NULL;
#if defined(PSP) && defined(EXPERIMENTAL)
static void *fbp0 = NULL, *dList;
#endif
static void _atexit() {
	set_clock(initial_clock);
	#if defined(PSP) && defined(EXPERIMENTAL)
	sceGuTerm();
	if(dList) free(dList);
	#endif
	SDL_Quit();
}

inline void sdl_error() {
	char *e = SDL_GetError();
	say("SDL Error: %s", e);
	next_line();
}
void init_video(Uint16 w, Uint16 h) {
	static char initd = 0;
	if(initd) return;
	#if defined(PSP) && defined(EXPERIMENTAL)
	device scr_acc = access_device(access_screen);
	scr_acc.screen->w = w;
	scr_acc.screen->h = h;
	dList = memalign(16, 262144);
	sceGuInit();
	sceGuStart(GU_DIRECT, dList);
	sceGuDrawBuffer(GU_PSM_8888, fbp0, 512);
	sceGuDispBuffer(w, h, (void*)0x88000, 512);
	sceGuDepthBuffer((void*)0x110000, 512);
	sceGuOffset(2048 - (w/2), 2048 - (h/2));
	sceGuViewport(2048, 2048, w, h);
	sceGuDepthRange(65535, 0);
	sceGuScissor(0, 0, w, h);
	sceGuEnable(GU_SCISSOR_TEST);
	sceGuFrontFace(GU_CW);
	sceGuShadeModel(GU_SMOOTH);
	sceGuEnable(GU_TEXTURE_2D);
	sceGuDepthFunc(GU_ALWAYS);
	sceGuClearColor(GU_COLOR(0.5f, 0.5f, 0.5f, 1.0f));
	sceGuClearDepth(0);
	sceGuFinish();
	reset_gu();
	sceGuSync(0,0);
	sceDisplayWaitVblankStart();
	sceGuDisplay(GU_TRUE);
	//reset_gu();
	clear_screen();
	flip_screen(1);
	#else
	SDL_ShowCursor(SDL_DISABLE);
	SDL_Surface *screen = SDL_SetVideoMode(w,h,0,SDL_HWSURFACE|SDL_DOUBLEBUF);
	if(!screen) {
		sdl_error();
	}
	#endif
}
#if defined(PSP) && defined(EXPERIMENTAL)
void reset_gu() {
	device scr_acc = access_device(access_screen);
	SDL_Surface *screen = scr_acc.screen;
	sceGuStart(GU_DIRECT,dList);	
	sceGumMatrixMode(GU_PROJECTION);
	sceGumLoadIdentity();
	sceGumOrtho(0,screen->w,screen->h,0,-1,1);
	sceGumMatrixMode(GU_VIEW);
	sceGumLoadIdentity();
	sceGumMatrixMode(GU_MODEL);
	sceGumLoadIdentity();
	sceGuTexMode(GU_PSM_8888, 0, 0, 0);	
	sceGuTexFunc(GU_TFX_REPLACE, GU_TCC_RGBA);
	sceGuTexFilter(GU_LINEAR, GU_LINEAR);
	sceGuTexWrap(GU_CLAMP,GU_CLAMP);
	sceGuFinish();
}
#endif
device access_device(int device_code) {
	static SDL_Surface *screen = NULL;
	static SDL_Joystick *joy = NULL;
	device ret;
	ret.type = -1;
	ret.screen = NULL;
	ret.joystick = NULL;
	switch(device_code) {
		case access_screen:
			#if defined(PSP) && defined(EXPERIMENTAL)
			if(!screen) screen = SDL_CreateRGBSurfaceFrom(
				(void*)(((unsigned int)fbp0)+0x44000000),
				480, 272, 32, 512*4,
				0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000
			);
			if(screen->pixels != (void*)(((unsigned int)fbp0)+0x44000000)) {
				screen->pixels = (void*)(((unsigned int)fbp0)+0x44000000);
			}
			#else
			if(!screen) screen = SDL_GetVideoSurface();
			#endif
			ret.screen = screen;
			ret.type = access_screen;
		break;
		case access_joystick:
			if(!joy) {
				if(SDL_JoystickOpened(0)) return ret;
				joy = SDL_JoystickOpen(0);
			}
			ret.joystick = joy;
			ret.type = access_joystick;
		break;
	}
	return ret;
}
int* access_int_global(int global_code) {
	static int 
		_resize_method = resize_nn,
		_zoom_persist = 0,
		_rotate_persist = 0,
		_pan_rate = 20,
		_scroll_skip = 5,
		_precaching = 0,
		_manga_mode = 0,
		_zoom_box_w = 150,
		_zoom_box_h = 200,
		_bookmark_on_load = 0,
		_fullview_mode = 0;
	switch(global_code) {
		case access_resize: return &_resize_method;
		case access_pan_rate: return &_pan_rate;
		case access_zoom_persist: return &_zoom_persist;
		case access_rotate_persist: return &_rotate_persist;
		case access_scroll_skip: return &_scroll_skip;
		case access_precaching: return &_precaching;
		case access_manga_mode: return &_manga_mode;
		case access_zoom_box_w: return &_zoom_box_w;
		case access_zoom_box_h: return &_zoom_box_h;
		case access_bookmark_on_load: return &_bookmark_on_load;
		case access_fullview_mode: return &_fullview_mode;
	}
	return NULL;
}
int set_clock(int new_clock) {
	#ifdef PSP
	if((new_clock > 333) || (new_clock < 20)) return 0;
	return scePowerSetClockFrequency(new_clock, new_clock, new_clock/2);
	#else
	return 0;
	#endif
}
int init_zoom_box(void *zbi) {
	zoom_box_info *zbid = (zoom_box_info*)zbi;
	SDL_Delay(2000);
	if(*(zbid->action_status) != zoom_in) return 0;
	if(
		(zbid->scaled_page->w < *access_int_global(access_zoom_box_w)) ||
		(zbid->scaled_page->h < *access_int_global(access_zoom_box_h))
	) return 0; //Please do not remove this line. The zoom box gets very broken
				//without this check. Yes, this check does disable the zoom box
				//for small surfaces. This is intentional to keep it from
				//breaking.
	show_image(zbid->scaled_page,*zbid->offsetX,*zbid->offsetY,0);
	draw_zoom_box(zbid->page_img,zbid->scale_ratio,zbid->rotation,-1,-1,*zbid->offsetX,*zbid->offsetY,2);
	zbid->zoom_box_is_active = 1;
	return 1;
}
void flip_screen(char flip_text) {
	#if defined(PSP) && defined(EXPERIMENTAL)
	if(flip_text) pspDebugScreenSetOffset((int)fbp0);
	sceDisplayWaitVblankStart();
	fbp0 = sceGuSwapBuffers();
	if(!flip_text) pspDebugScreenSetOffset((int)fbp0);
	#else
	device scr_acc = access_device(access_screen);
	#ifdef PSP
	if(flip_text)
		pspDebugScreenSetOffset((int)scr_acc.screen->pixels & 0x1FFFFF);
	#endif
	SDL_Flip(scr_acc.screen);
	#ifdef PSP
	if(!flip_text)
		pspDebugScreenSetOffset((int)scr_acc.screen->pixels & 0x1FFFFF);
	#endif
	#endif
}
void clear_screen() {
	#if defined(PSP) && defined(EXPERIMENTAL)
	sceGuStart(GU_DIRECT,dList);
	sceGuClear(GU_COLOR_BUFFER_BIT|GU_DEPTH_BUFFER_BIT);
	sceGuFinish();
	sceGuSync(0,0);
	#else
	device scr_acc = access_device(access_screen);
	SDL_FillRect(scr_acc.screen,NULL,SDL_MapRGB(scr_acc.screen->format,127,127,127));
	#endif
}

char show_image(SDL_Surface *surf, Uint16 offsetX, Uint16 offsetY, char flip) {
	if(!surf) return 0;
	SDL_Rect dest, source;
	device scr_acc = access_device(access_screen);
	SDL_Surface *screen = scr_acc.screen;
	dest.w = screen->w;
	dest.h = screen->h;
	if(surf->w < screen->w) dest.x = (screen->w-surf->w)/2;
	else dest.x = 0;
	if(surf->h < screen->h) dest.y = (screen->h-surf->h)/2;
	else dest.y = 0;
	source.w = (screen->w > surf->w)?screen->w:surf->w;
	source.h = (screen->h > surf->h)?screen->h:surf->h;
	source.x = offsetX;
	source.y = offsetY;
	#if defined(PSP) && defined(EXPERIMENTAL)
	SDL_Surface *new_surf = SDL_CreateRGBSurfaceFrom(
		//memalign(16,512*source.h*4),
		//source.w, source.h, 32, 512*4,
		(void*)(((unsigned int)fbp0)+0x4000000),
		screen->w, screen->h, 32, 512*4,
		0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000
	);
	if(!new_surf) return 0;
	clear_screen();
	{
		SDL_Rect temp = {0,0,0,0};
		if(SDL_BlitSurface(surf,&source,new_surf,&temp) == -1) return 0;
	}
	//sceKernelDcacheWritebackInvalidateAll();
	//sceGuStart(GU_DIRECT,dList);
	//sceGuCopyImage(GU_PSM_8888,0,0,source.w,source.h,512,new_surf->pixels,0,0,512,(void*)(((unsigned int)fbp0)+0x4000000));
	//sceGuFinish();
	sceGuSync(0,0);
	if(flip) flip_screen(1);
	free(new_surf->pixels);
	SDL_FreeSurface(new_surf);
	return 1;
	//return show_image2(surf,&source,NULL,0,flip);
	#else
	if(surf->flags & SDL_SRCALPHA) {
		clear_screen();
		//SDL_SetAlpha(surf,0,SDL_ALPHA_OPAQUE);
	}
	if(SDL_BlitSurface(surf,&source,screen,&dest) > -1) {
		if(flip) flip_screen(1);
		return 1;
	} else {
		sdl_error();
		return 0;
	}
	#endif
}
char show_image2(SDL_Surface *surf, SDL_Rect *src, SDL_Rect *dst, char rotation, char flip) {
	device scr_acc = access_device(access_screen);
	SDL_Surface *screen = scr_acc.screen;
	#if defined(PSP) && defined(EXPERIMENTAL)
	SDL_Surface *clipped = SDL_CreateRGBSurfaceFrom(memalign(16,256*256*4),
		256, 256, 32, 256*4,
		0x000000FF, 0x0000FF00, 0x00FF0000, 0xFF000000
	);
	SDL_Rect clipper, temp = {0,0,0,0},
		temp_src = {0,0,0,0}, temp_dst = {0,0,0,0};
	if(!src) {
		src = &temp_src;
		src->w = surf->w;
		src->h = surf->h;
	}
	if(!dst) {
		dst = &temp_dst;
		dst->w = src->w;
		dst->h = src->h;
	}
	clipper.w = 256;
	clipper.h = 256;
	sceGuStart(GU_DIRECT,dList);
	sceGuClear(GU_COLOR_BUFFER_BIT|GU_DEPTH_BUFFER_BIT);
	sceGuTexScale(1.0f, 1.0f);
	sceGuTexOffset(0.0f, 0.0f);
	sceGuTexImage(0, 256, 256, 256, clipped->pixels);
	vertexel16 *corners = sceGuGetMemory(sizeof(vertexel16)*2);
	memset(corners,0,sizeof(vertexel16)*2);
	corners[1].x = corners[1].y = 256;
	corners[1].u = corners[1].v = 256;
	sceGuFinish();
	sceGuSync(0,0);
	int i = 0, j = 0;
	for(i = dst->x; i <= dst->x+dst->w; i += 256) {
		clipper.x = i+src->x;
		corners[0].x = i;
		corners[1].x = i+256;
		for(j = dst->y; j <= dst->y+dst->h; j += 256) {
			clipper.y = j+src->y;
			corners[0].y = j;
			corners[1].y = j+256;
			SDL_FillRect(clipped,NULL,SDL_MapRGBA(clipped->format,0,0,0,SDL_ALPHA_TRANSPARENT));
			if(SDL_BlitSurface(surf,&clipper,clipped,&temp) == -1) return -2;
			sceKernelDcacheWritebackInvalidateAll();
			sceGuStart(GU_DIRECT,dList);
			sceGumDrawArray(GU_SPRITES, GU_TEXTURE_16BIT|GU_VERTEX_32BITF|GU_TRANSFORM_2D, 2, NULL, corners);
			sceGuFinish();
			sceGuSync(0,0);
		}
	}
	if(flip) {
		flip_screen(1);
	}
	//free(corners);
	free(clipped->pixels);
	SDL_FreeSurface(clipped);
	return 1;
	#else
	SDL_Rect src_temp = {0,0,0,0}, dst_temp = {0,0,0,0};
	SDL_Surface *buffer1, *buffer2;
	if(!src) {
		src_temp.w = surf->w;
		src_temp.h = surf->h;
		src = &src_temp;
		buffer1 = surf;
	} else {
		buffer1 = SDL_CreateRGBSurface(
			SDL_SWSURFACE, src->w, src->h, 32,
			0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF
		);
		if(!buffer1) return 0;
		SDL_BlitSurface(surf,src,buffer1,&dst_temp);
	}
	if(!dst) {
		dst_temp.w = screen->w;
		dst_temp.h = screen->h;
	} else dst_temp = *dst;
	dst = &dst_temp;
	buffer2 = (rotation & 1?
		rotozoom(buffer1,NULL,dst->h,dst->w,rotation):
		rotozoom(buffer1,NULL,dst->w,dst->h,rotation)
	);
	if((buffer1 != buffer2) && (buffer1 != surf)) SDL_FreeSurface(buffer1);
	if(!buffer2) return 0;
	SDL_SetAlpha(buffer2,0,SDL_ALPHA_OPAQUE);
	SDL_Rect temp = {0,0,buffer2->w,buffer2->h};
	SDL_BlitSurface(buffer2,&temp,screen,dst);
	if(buffer2 != surf) SDL_FreeSurface(buffer2);
	return 1;
	#endif
}
SDL_Surface* load_background() {
	static SDL_Surface* bg = NULL;
	static char loaded = 0;
	device scr_acc = access_device(access_screen);
	SDL_Surface *screen = scr_acc.screen;
	if(!loaded) {
		SDL_Surface *buffer = IMG_Load("background.png");
		if(!buffer) return NULL;
		bg = rotozoom(buffer,NULL,screen->w,screen->h,0);
		if(bg != buffer) SDL_FreeSurface(buffer);
		loaded = 1;
	}
	if(bg) return bg;
	return NULL;
}
event show_book(char *dir, char *book_name) {
	setXY(0,0);
	say("Loading...");
	comic_book book = open_comic_book(dir,book_name);
	if(book.num_pages == 0) {
		say("Error opening book.");
		next_line();
		return init_event(error_ev);
	}
	event operation = init_event(no_event);
	float pscale_ratio = (*access_int_global(access_fullview_mode)?1.0f:0.0f);
	Uint8 protation = 0;
	if(*access_int_global(access_bookmark_on_load))
		book = goto_page(book,get_bookmark(book.filename));
	SDL_Surface *page_img;
	precache_info pci = {NULL,0,0,&book};
	SDL_Thread *precacher;
	while(!(operation.command & (quit_command|close_book|open_book|error_ev))) {
		if(!*access_int_global(access_rotate_persist)) protation = 0;
		if(!*access_int_global(access_zoom_persist))
			pscale_ratio = (*access_int_global(access_fullview_mode)?1.0f:0.0f);
		if((pci.page == book.current_page) && pci.new_page && (pci.loaded == 2)) {
			page_img = pci.new_page;
		} else {
			if(pci.loaded == 1) {
				SDL_WaitThread(precacher,NULL);
			}
			if((pci.loaded == 2) && (pci.page != book.current_page))
				SDL_FreeSurface(pci.new_page);
			int ret;
			if((pci.loaded == 1) && (pci.page == book.current_page))
				SDL_WaitThread(precacher,&ret);
			if((pci.loaded == 2) && (pci.page == book.current_page)) {
				 page_img = pci.new_page;
			} else {
				size_t size = 0;
				char *block = extract_file(&book,book.current_page,&size);
				if(block == NULL) {
					Uint16 page = book.current_page;
					close_comic_book(book);
					book = open_comic_book(dir,book_name);
					if(book.num_pages == 0) {
						say("Error opening book.");
						next_line();
						return init_event(error_ev);
					}
					book.current_page = page;
					block = extract_file(&book,book.current_page,&size);
					if(block == NULL) {
						say("Could not load page from archive. Bailing...");
						next_line();
						return init_event(error_ev);
					}
				}
				SDL_RWops *page = SDL_RWFromMem(block,size);
				if(page) {
					page_img = IMG_Load_RW(page,0);
					SDL_FreeRW(page);
				}
				free(block);
				free_event(operation);
			}
		}
		pci.loaded = 0;
		pci.page = book.current_page + 1;
		if(*access_int_global(access_precaching) && (pci.page < book.num_pages)) precacher = SDL_CreateThread(cache_page,&pci);
		if(!page_img) {
			clear_screen();
			setXY(0,0);
			say("SDL_image error: %s",IMG_GetError());
			next_line();
			operation.command = error_ev;
		} else {
			operation = show_page(page_img,&book,&protation,&pscale_ratio);
			SDL_FreeSurface(page_img);
		}
		if(operation.command == error_ev) {
			event op2 = init_event(no_event);
			while(op2.command == no_event) {
				op2.command = handle_input(1);
				if(op2.command == open_menu) {
					free_event(op2);
					op2 = show_menu(&book);
				}
				switch(op2.command) {
					case quit_command: case next_page: case prev_page:
					case jump_page: case close_book: case open_book: break;
					default: op2.command = no_event;
				}
			}
			operation = op2;
		}
		switch(operation.command) {
			case next_page: book = goto_next_page(book); break;
			case prev_page: book = goto_prev_page(book); break;
			case jump_page: book = goto_page(book, operation.int_data); break;
		}
	}
	++pci.page;
	close_comic_book(book);
	return operation;
}
int cache_page(void *pci) {
	#ifdef PSP
	SceKernelThreadInfo info;
	SceUID thid = sceKernelGetThreadId();
	info.size = sizeof(SceKernelThreadInfo);
	if(!sceKernelReferThreadStatus(thid,&info))
		sceKernelChangeThreadPriority(thid,info.currentPriority+8);
	#endif
	precache_info *pcis = (precache_info*)pci;
	pcis->loaded = 1;
	if(pcis->page >= pcis->book->num_pages) return 0;
	Uint16 precache_page = pcis->page;
	size_t size = 0;
	char *block = extract_file(pcis->book,pcis->page,&size);
	if(block == NULL) {
		pcis->loaded = -2;
		return -2;
		//Don't try any fancy stuff here.
		//If you suspend your PSP while it's try to precache, just bail.
	}
	SDL_RWops *page = SDL_RWFromMem(block,size);
	if(page) {
		pcis->new_page = IMG_Load_RW(page,0);
		SDL_FreeRW(page);
	} else {
		pcis->loaded = -3;
		return -3;
	}
	if(!pcis->new_page) {
		pcis->loaded = -3;
		return -3;
	}
	free(block);
	if(pcis->page != precache_page) {
		SDL_FreeSurface(pcis->new_page);
		pcis->loaded = -1;
		return -1;
	}
	pcis->loaded = 2;
	return 1;
}

inline void* access_pixel(const SDL_Surface* const surf, const Uint16 x, const Uint16 y) {
	return (void*)((char*)surf->pixels + x*surf->format->BytesPerPixel + surf->pitch * y);
}
inline void copy_pixel(void const *src, void *dst, Uint8 bpp) {
	#ifdef PSP
	/*asm(
		"loop:				\n"
		"	lb	$t0, 0(%1)	\n"
		"	sb	$t0, 0(%0)	\n"
		"	addi %0, 1		\n"
		"	addi %1, 1		\n"
		"	addi %2, -1		\n"
		"	bgtz %2, loop"
		: : "r" (dst), "r" (src), "r" (bpp)
	);*/ //Doesn't work when inlined, too slow when not inlined
	while(bpp--) *((char*)dst++) = *((char*)src++);
	#else
	memcpy(dst,src,bpp);
	#endif
}
inline void blend(void *pixel, SDL_PixelFormat *fmt, Uint8 r, Uint8 g, Uint8 b, Uint8 a) {
	Uint8 old_r, old_g, old_b, old_a;
	Uint8 new_r, new_g, new_b, new_a;
	switch(fmt->BytesPerPixel) {
		case 1:
			SDL_GetRGBA(*(Uint8*)pixel,fmt,&old_r,&old_g,&old_b,&old_a);
		break;
		case 2:
			SDL_GetRGBA(*(Uint16*)pixel,fmt,&old_r,&old_g,&old_b,&old_a);
		break;
		case 3: {
			Uint32 buffer;
			buffer = ((Uint8*)pixel)[0] | (((Uint8*)pixel)[1] << 8) | (((Uint8*)pixel)[2] << 16);
			SDL_GetRGBA(buffer,fmt,&old_r,&old_g,&old_b,&old_a);
		} break;
		case 4:
			SDL_GetRGBA(*(Uint32*)pixel,fmt,&old_r,&old_g,&old_b,&old_a);
		break;
	}
	new_r = (Uint8) ((((Uint32)old_r << 8) + ((Uint32)r * a)) / (256 + a));
	new_g = (Uint8) ((((Uint32)old_g << 8) + ((Uint32)g * a)) / (256 + a));
	new_b = (Uint8) ((((Uint32)old_b << 8) + ((Uint32)b * a)) / (256 + a));
	new_a = a + old_a;
	Uint32 new_color = SDL_MapRGBA(fmt,new_r,new_g,new_b,new_a);
	switch(fmt->BytesPerPixel) {
		case 1: *(Uint8*)pixel = (Uint8)new_color; break;
		case 2: *(Uint16*)pixel = (Uint16)new_color; break;
		case 3: {
			((Uint8*)pixel)[0] = new_color;
			((Uint8*)pixel)[1] = new_color >> 8;
			((Uint8*)pixel)[2] = new_color >> 16;
		} break;
		case 4: *(Uint32*)pixel = (Uint32)new_color; break;
	}
}

#ifndef USE_GFX
SDL_Surface* resize(SDL_Surface *old, SDL_Rect *clip, Uint16 newW, Uint16 newH) {
	SDL_Rect src_rect;
	if(clip == NULL) {
		src_rect.x = 0;
		src_rect.y = 0;
		src_rect.w = old->w;
		src_rect.h = old->h;
	} else src_rect = *clip;
	if((newW == 0) && (newH != 0)) newW = (Uint16)(((float)newH/src_rect.h)*old->w);
	else if((newH == 0) && (newW != 0)) newH = (Uint16)(((float)newW/src_rect.w)*old->h);
	else if((newW == 0) && (newH == 0)) return NULL;
	if((clip == NULL) && (newW == old->w) && (newH == old->h)) return old;
	int resize_method = *access_int_global(access_resize);
	if((newW/old->w >= 2) && (newH/old->h >= 2)) resize_method = resize_nn;
	SDL_Surface *new_surf = SDL_CreateRGBSurface(
		SDL_SWSURFACE,
		newW, newH,
		((resize_method == resize_nn) || (old->format->BytesPerPixel >= 3)) ? old->format->BitsPerPixel : 24,
		old->format->Rmask, old->format->Gmask, old->format->Bmask, old->format->Amask
	);
	if(!new_surf ||
	  ((new_surf->format->BytesPerPixel == 1) && !(copy_palette(old, new_surf) == 1))
	) {
		sdl_error();
		return NULL;
	}
	SDL_LockSurface(old);
	SDL_LockSurface(new_surf);
	Uint16 i, j;
	void *old_pixel, *new_pixel;
	float locationX, locationY;
	switch(resize_method) {
		case resize_nn: 
			for(i = 0; i < newW; ++i) {
				for(j = 0; j < newH; ++j) {
					locationX = ((float)i)/newW*src_rect.w + src_rect.x;
					locationY = ((float)j)/newH*src_rect.h + src_rect.y;
					old_pixel = access_pixel(old,(Uint16)(locationX),(Uint16)(locationY));
					new_pixel = access_pixel(new_surf,i,j);
					copy_pixel(old_pixel,new_pixel,old->format->BytesPerPixel);
				}
			}
		break;
		case resample: {
			Uint16 cornerX0, cornerX1, cornerY0, cornerY1;
			Uint8 *old_pixel0, *old_pixel1;
			SDL_Surface *buffer = NULL;
			if(old->format->BytesPerPixel < 3) {
				buffer = old;
				old = SDL_CreateRGBSurface(
					SDL_SWSURFACE,
					buffer->w, old->h, 24,
					buffer->format->Rmask, buffer->format->Gmask,
					buffer->format->Bmask, buffer->format->Amask
				);
				SDL_UnlockSurface(buffer);
				SDL_BlitSurface(buffer,NULL,old,NULL);
				SDL_UpdateRect(old,0,0,0,0);
				SDL_LockSurface(old);
			}
			Uint8 k = new_surf->format->BytesPerPixel;
			Uint16 val;
			for(i = newW; i--;) {
				for(j = newH; j--;) {
					locationX = ((float)i)/newW*(src_rect.w-1) + src_rect.x;
					locationY = ((float)j)/newH*(src_rect.h-1) + src_rect.y;
					new_pixel = access_pixel(new_surf,i,j);
					cornerX0 = (Uint16)(locationX);
					cornerY0 = (Uint16)(locationY);
					cornerX1 = ceil(locationX);
					cornerY1 = ceil(locationY);
					old_pixel0 = access_pixel(old,cornerX0,cornerY0);
					old_pixel1 = access_pixel(old,cornerX1,cornerY1);
					val = ((Uint16)(*(old_pixel0))+
						(Uint16)(*(old_pixel1)))>>1;
					*((Uint8*)new_pixel) = (Uint8)val;
					val = ((Uint16)(*(++old_pixel0))+
						(Uint16)(*(++old_pixel1)))>>1;
					*((Uint8*)++new_pixel) = (Uint8)val;
					val = ((Uint16)(*(++old_pixel0))+
						(Uint16)(*(++old_pixel1)))>>1;
					*((Uint8*)++new_pixel) = (Uint8)val;
					if(k == 4) {
						val = ((Uint16)(*(++old_pixel0))+
							(Uint16)(*(++old_pixel1)))>>1;
						*((Uint8*)++new_pixel) = (Uint8)val;
					}
				}
			}
			if(buffer) SDL_FreeSurface(old);
		} break;
	}
	SDL_UnlockSurface(old);
	SDL_UnlockSurface(new_surf);
	return new_surf;
}
SDL_Surface* rotate(SDL_Surface *old, char amount) {
	while(amount < 0) amount += 4;
	amount %= 4;
	if(amount == 0) return old;
	SDL_Surface *new_surf = SDL_CreateRGBSurface(
		SDL_SWSURFACE,
		(amount & 1?old->h:old->w), (amount & 1?old->w:old->h),
		old->format->BitsPerPixel,
		old->format->Rmask, old->format->Gmask, old->format->Bmask, old->format->Amask
	);
	if(!new_surf ||
	  ((new_surf->format->BytesPerPixel == 1) && !(copy_palette(old, new_surf) == 1))
	) {
		sdl_error();
		return NULL;
	} else if(new_surf->format->BytesPerPixel == 1) copy_palette(old, new_surf);
	SDL_LockSurface(old);
	SDL_LockSurface(new_surf);
	Uint16 i, j;
	void *old_pixel, *new_pixel;
	for(i = 0; i < new_surf->w; i++) {
		for(j = 0; j < new_surf->h; j++) {
			if(amount == 1) old_pixel = access_pixel(old,j,old->h-i-1);
			else if(amount == 2) old_pixel = access_pixel(old,old->w-i-1,old->h-j-1);
			else if(amount == 3) old_pixel = access_pixel(old,old->w-j-1,i);
			new_pixel = access_pixel(new_surf,i,j);
			copy_pixel(old_pixel,new_pixel,old->format->BytesPerPixel);
		}
	}
	SDL_UnlockSurface(old);
	SDL_UnlockSurface(new_surf);
	return new_surf;
}
#endif
SDL_Surface* rotozoom(SDL_Surface *old, SDL_Rect *clip, Uint16 newW, Uint16 newH, const char rotation) {
	#ifdef USE_GFX
	float ratioW = 0, ratioH = 0;
	SDL_Surface *buffer;
	if(clip) {
		SDL_Rect temp = {0,0,0,0};
		buffer = SDL_CreateRGBSurface(
			SDL_SWSURFACE,clip->w,clip->h,32,
			0xFF000000,0x00FF0000,0x0000FF00,0x000000FF
		);
		SDL_BlitSurface(old,clip,buffer,&temp);
	} else buffer = old;
	if((newW == 0) && (newH != 0)) ratioW = ratioH = (float)newH/buffer->h;
	else if((newH == 0) && (newW != 0)) ratioH = ratioW = (float)newW/buffer->w;
	else if((newW == 0) && (newH == 0)) return NULL;
	else {
		ratioW = (float)newW/buffer->w;
		ratioH = (float)newH/buffer->h;
	}
	if(old->format->BytesPerPixel == 1) {
		buffer = SDL_CreateRGBSurface(
			SDL_SWSURFACE,old->w,old->h,32,
			0xFF000000,0x00FF0000,0x0000FF00,0x000000FF
		);
		SDL_BlitSurface(old,NULL,buffer,NULL);
	}
	if((clip == NULL) && (newW == old->w) && (newH == old->h) && !rotation) return old;
	SDL_Surface *new_surf = rotozoomSurfaceXY(
		buffer,-rotation*90,
		ratioW,ratioH,*access_int_global(access_resize)
	);
	if((new_surf != buffer) && (buffer != old)) SDL_FreeSurface(buffer);
	if(!new_surf) return NULL;
	return new_surf;
	#else
	SDL_Surface *buffer = resize(old,clip,newW,newH);
	if(!buffer) return NULL;
	SDL_Surface *buffer2 = rotate(buffer,rotation);
	if((buffer2 != buffer) && (buffer != old)) SDL_FreeSurface(buffer);
	if(!buffer2) return NULL;
	return buffer2;
	#endif
}
int copy_palette(SDL_Surface *old, SDL_Surface *new_surf) {
	if((old->format->BytesPerPixel != 1) || (new_surf->format->BytesPerPixel != 1))
		return -1;
	SDL_Palette *pal = old->format->palette;
	SDL_Color *colors = malloc(sizeof(SDL_Color)*pal->ncolors);
	if(!colors) {
		say("Out of memory");
		next_line();
		return -2;
	}
	memcpy(colors,pal->colors,pal->ncolors*sizeof(SDL_Color));
	return SDL_SetPalette(new_surf, SDL_LOGPAL|SDL_PHYSPAL, colors, 0, pal->ncolors);
}
void blend_rect(SDL_Surface *surf, SDL_Rect rect, Uint8 r, Uint8 g, Uint8 b, Uint8 a) {
	if((rect.w == 0) || (rect.h == 0)) return;
	SDL_LockSurface(surf);
	Uint16 i;
	void *pixel;
	for(i = 0; i < rect.w; i++) {
		pixel = access_pixel(surf,rect.x+i,rect.y);
		blend(pixel,surf->format,r,g,b,a);
	}
	if(rect.h > 1) {
		for(i = 1; i < rect.h; i++) {
			pixel = access_pixel(surf,rect.x,rect.y+i);
			blend(pixel,surf->format,r,g,b,a);
		}
		if(rect.w > 1) {
			for(i = 1; i < rect.h; i++) {
				pixel = access_pixel(surf,rect.x+rect.w-1,rect.y+i);
			blend(pixel,surf->format,r,g,b,a);
			}
			for(i = 1; i < rect.w-1; i++) {
				pixel = access_pixel(surf,rect.x+i,rect.y+rect.h-1);
				blend(pixel,surf->format,r,g,b,a);
			}
		}
	}
	SDL_UnlockSurface(surf);
}
void draw_zoom_box(SDL_Surface *page_img, float scale_ratio, Uint8 rotation, Sint32 x, Sint32 y, Uint16 offsetX, Uint16 offsetY, Uint8 zoom_level) {
	device scr_acc = access_device(access_screen);
	SDL_Surface *screen = scr_acc.screen;
	SDL_Rect big_rect, little_rect, src_rect, dst_rect;
	big_rect.w = *access_int_global(access_zoom_box_w);
	big_rect.h = *access_int_global(access_zoom_box_h); 
	float newW, newH;
	newW = (rotation & 1?page_img->h*scale_ratio:page_img->w*scale_ratio);
	newH = (rotation & 1?page_img->w*scale_ratio:page_img->h*scale_ratio);
	Uint16 smallerW, smallerH;
	smallerW = (screen->w > newW?newW:screen->w);
	smallerH = (screen->h > newH?newH:screen->h);
	if(zoom_level > 4) zoom_level = 4;
	else if(zoom_level == 0) zoom_level = 1;
	little_rect.w = big_rect.w/zoom_level;
	if(little_rect.w > newW) little_rect.w = newW;
	little_rect.h = big_rect.h/zoom_level;
	if(little_rect.h > newH) little_rect.h = newH;
	if(x > (Uint16)newW) little_rect.x = 0;
	else if(x < 0) little_rect.x = (smallerW - little_rect.w)/2;
	else little_rect.x = x;
	if(y > (Uint16)newH) little_rect.y = 0;
	if(y < 0) little_rect.y = (smallerH - little_rect.h)/2;
	else little_rect.y = y;
	if(little_rect.x + little_rect.w/2 > smallerW/2) {
		big_rect.x = 10;
		big_rect.y = (screen->h - big_rect.h)/2;
	} else {
		big_rect.x = screen->w - big_rect.w - 10;
		big_rect.y = (screen->h - big_rect.h)/2;
	}
	dst_rect = big_rect;
	dst_rect.x++;
	dst_rect.y++;
	dst_rect.w -= 2;
	dst_rect.h -= 2;

	switch(rotation % 4) {
		case 0:
			src_rect.x = (Uint16)((little_rect.x + offsetX)/scale_ratio);
			src_rect.y = (Uint16)((little_rect.y + offsetY)/scale_ratio);
			src_rect.w = (Uint16)(little_rect.w/scale_ratio);
			src_rect.h = (Uint16)(little_rect.h/scale_ratio);
		break;
		case 1:
			src_rect.h = (Uint16)(little_rect.w/scale_ratio);
			src_rect.w = (Uint16)(little_rect.h/scale_ratio);
			src_rect.x = (Uint16)((little_rect.y + offsetY)/scale_ratio);
			src_rect.y = page_img->h - src_rect.h - (Uint16)((little_rect.x + offsetX)/scale_ratio);
		break;
		case 2:
			src_rect.w = (Uint16)(little_rect.w/scale_ratio);
			src_rect.h = (Uint16)(little_rect.h/scale_ratio);
			src_rect.x = page_img->w - src_rect.w - (Uint16)((little_rect.x + offsetX)/scale_ratio);
			src_rect.y = page_img->h - src_rect.h - (Uint16)((little_rect.y + offsetY)/scale_ratio);
		break;
		case 3:
			src_rect.h = (Uint16)(little_rect.w/scale_ratio);
			src_rect.w = (Uint16)(little_rect.h/scale_ratio);
			src_rect.x = page_img->w - src_rect.w - (Uint16)((little_rect.y + offsetY)/scale_ratio);
			src_rect.y = (Uint16)((little_rect.x + offsetX)/scale_ratio);
		break;
	}

	show_image2(page_img,&src_rect,&dst_rect,rotation,0);
	src_rect.x = src_rect.y = 0;
	src_rect.w = dst_rect.w;
	src_rect.h = dst_rect.h;
	{ //More stack space saving!
		float visiW, visiH;
		visiW = (rotation & 1?scale_ratio*page_img->h:scale_ratio*page_img->w);
		visiH = (rotation & 1?scale_ratio*page_img->w:scale_ratio*page_img->h);
		if(screen->w > visiW)
			little_rect.x += (screen->w - visiW)/2;
		if(screen->h > visiH)
			little_rect.y += (screen->h - visiH)/2;
	}
	#ifdef USE_GFX
	rectangleRGBA(
		screen, little_rect.x, little_rect.y,
		little_rect.x+little_rect.w-1, little_rect.y+little_rect.h-1,
		95, 95, 255, 191
	);
	rectangleRGBA(
		screen, big_rect.x, big_rect.y,
		big_rect.x+big_rect.w-1, big_rect.y+big_rect.h-1,
		191, 63, 63, 127
	);
	#else
	blend_rect(screen, little_rect, 95, 95, 255, 191);
	blend_rect(screen, big_rect, 191, 63, 63, 127);
	#endif
	flip_screen(1);
}

int handle_input(char wait) {
	static int down = 0;
	SDL_Event ev;
	device joy_acc = access_device(access_joystick);
	SDL_Joystick *joy = joy_acc.joystick;
	if(wait > 0?SDL_WaitEvent(&ev):SDL_PollEvent(&ev)) {
		switch(ev.type) {
			case SDL_JOYBUTTONDOWN: case SDL_JOYBUTTONUP: {
				int ret;
				switch(ev.jbutton.button) {
					case 8:	ret = pan_up; break;
					case 9:	ret = pan_right; break;
					case 6:	ret = pan_down; break;
					case 7:	ret = pan_left; break;
					case 4:	ret = prev_page; break;
					case 5:	ret = next_page; break;
					case 3:	ret = rotate_ccw; break;
					case 1:	ret = rotate_cw; break;
					case 0:	ret = zoom_in; break;
					case 2:	ret = zoom_out; break;
					case 10: ret = zoom_fixed; break;
					case 11: ret = open_menu; break;
					default: ret = no_event; break;
				}
				if((!SDL_JoystickGetButton(joy,ev.jbutton.button)) && (ret != no_event)) {
					if(down & ret) down ^= ret;
					ret |= event_end;
					return ret;
				}
				if(down & ret) return no_event;
				down |= ret;
				return ret;
			} break;
			case SDL_JOYAXISMOTION: {
				return pan_event;
			} break;
			#ifndef PSP
			case SDL_QUIT: return quit_command;
			case SDL_KEYDOWN: case SDL_KEYUP: {
				int ret;
				SDL_KeyboardEvent k = ev.key;
				switch(k.keysym.sym) {
					case SDLK_UP:		ret = pan_up;	 break;
					case SDLK_RIGHT:	ret = pan_right; break;
					case SDLK_DOWN:		ret = pan_down;	 break;
					case SDLK_LEFT:		ret = pan_left;	 break;
					case SDLK_PAGEUP:	ret = prev_page; break;
					case SDLK_PAGEDOWN:	ret = next_page; break;
					case SDLK_PERIOD:	ret = rotate_ccw ;break;
					case SDLK_SLASH:	ret = rotate_cw; break;
					case SDLK_QUOTE:	ret = zoom_in;	 break;
					case SDLK_SEMICOLON:	ret = zoom_out; break;
					case SDLK_SPACE:	ret = open_menu; break;
					case SDLK_l: ret = zoom_fixed; break;
					case SDLK_q: ret = quit_command; break;
					default: ret = no_event; break;
				}
				if((k.state == SDL_RELEASED) && (ret ^ no_event)) ret |= event_end;
				return ret;
			} break;
			#endif
		}
	} else {
		if(wait > 0) sdl_error();
		return no_event;
	}
	return no_event;
}
event show_page(SDL_Surface *page_img, comic_book* book, Uint8 *rotation, float *scale_ratio) {
	setXY(0,0);
	if(!page_img) return init_event(error_ev);
	device scr_acc = access_device(access_screen);
	SDL_Surface *screen = scr_acc.screen;
	device joy_acc = access_device(access_joystick);
	SDL_Joystick *joy = joy_acc.joystick;
	Uint32 pan_amount = *access_int_global(access_pan_rate);
	Uint16 scale_width = (Uint16)(*scale_ratio?page_img->w**scale_ratio:screen->w);
	char joy_in_use = 0;
	/*SDL_Surface *new_surf = (*rotation & 1?
		resize(page_img,NULL,0,scale_width):
		resize(page_img,NULL,scale_width,0)
	);
	if(!new_surf) return init_event(error_ev);
	SDL_Surface *buffer = new_surf;
	new_surf = rotate(buffer,*rotation);
	if((new_surf != buffer) && (buffer != page_img)) SDL_FreeSurface(buffer);*/
	SDL_Surface *new_surf = rotozoom(page_img,NULL,scale_width,0,*rotation);
	if(!new_surf) return init_event(error_ev);
	clear_screen();
	Uint16 offsetX = 0, offsetY = 0;
	if(*access_int_global(access_manga_mode) && (new_surf->w > screen->w)) offsetX = new_surf->w - screen->w;
	show_image(new_surf,offsetX,0,1);
	setXY(0,33);
	say("%u/%u",book->current_page+1,book->num_pages);
	setXY(0,0);
	char wait_for_input = 1;
	unsigned short delay = 0;
	int quit = 0, poll_command, full_mode = (*scale_ratio?
		((*scale_ratio == 1)?full_view:custom_zoom):full_width);
	char offsetX_shift = 0, offsetY_shift = 0;
	event ret = init_event(no_event);
	clock_t start = 0;
	int pending_action = no_event;
	SDL_Thread *izb_thread = NULL;
	zoom_box_info zbi = {0,NULL,NULL,&offsetX,&offsetY,*scale_ratio,*rotation,NULL};
	Uint8 zb_zoom_level = 2;
	Uint16 zb_offsetX = 0;
	Uint16 zb_offsetY = 0;
	while(!quit) {
		poll_command = handle_input(wait_for_input);
		switch(poll_command) {
			case next_page: case prev_page:
			case rotate_cw: case rotate_ccw:
			case zoom_in: case zoom_out: case zoom_fixed:
			case open_menu:
				if(!pending_action) {
					pending_action = poll_command;
					start = clock();
				}
				if((pending_action == zoom_in) && !zbi.zoom_box_is_active) {
					zbi.action_status = &pending_action;
					zbi.page_img = page_img;
					zbi.scaled_page = new_surf;
					zbi.rotation = *rotation;
					zbi.scale_ratio = ((float)scale_width)/page_img->w;
					izb_thread = SDL_CreateThread(init_zoom_box,&zbi);
					zb_zoom_level = 2;
					zb_offsetX = (new_surf->w < *access_int_global(access_zoom_box_w))?0:
						(screen->w < new_surf->w)?
						(screen->w - *access_int_global(access_zoom_box_w)/zb_zoom_level)/2:
						(new_surf->w - *access_int_global(access_zoom_box_w)/zb_zoom_level)/2;
					zb_offsetY = (new_surf->h < *access_int_global(access_zoom_box_h))?0:
						(screen->h < new_surf->h)?
						(screen->h - *access_int_global(access_zoom_box_h)/zb_zoom_level)/2:
						(new_surf->h - *access_int_global(access_zoom_box_h)/zb_zoom_level)/2;
				}
			break;
			case next_page|event_end:
				if(pending_action != next_page) break;
				if((clock() - start)/CLOCKS_PER_SEC < 2) {
					if(book->current_page != book->num_pages - 1)
						quit = next_page;
				}
				pending_action = no_event;
			break;
			case prev_page|event_end:
				if(pending_action != prev_page) break;
				if((clock() - start)/CLOCKS_PER_SEC < 2) {
					if(book->current_page != 0)
						quit = prev_page;
				}
				pending_action = no_event;
			break;
			case quit_command|event_end: quit = quit_command; break;
			case pan_up:	offsetY_shift = -pan_amount;	break;
			case pan_down:	offsetY_shift =  pan_amount;	break;
			case pan_left:	offsetX_shift = -pan_amount;	break;
			case pan_right:	offsetX_shift =  pan_amount;	break;
			case rotate_cw|event_end: case rotate_ccw|event_end:
			if((pending_action != rotate_cw) && (pending_action != rotate_ccw)) break;
			if((clock() - start)/CLOCKS_PER_SEC < 2) {
				if(zbi.zoom_box_is_active) {
					{ //Begin new block...saving space on stack here.
						Uint16 centerX, centerY;
						Uint16 maxX =
							(screen->w > new_surf->w?new_surf->w:screen->w);
						Uint16 maxY =
							(screen->h > new_surf->h?new_surf->h:screen->h);
						Uint16 zb_w = *access_int_global(access_zoom_box_w);
						Uint16 zb_h = *access_int_global(access_zoom_box_h);
						centerX = zb_offsetX + zb_w/(zb_zoom_level*2);
						centerY = zb_offsetY + zb_h/(zb_zoom_level*2);
						zb_zoom_level += (pending_action == rotate_cw?-1:1);
						if(zb_zoom_level > 4) zb_zoom_level = 4;
						else if(zb_zoom_level < 2) zb_zoom_level = 2;
						if(zb_w/(zb_zoom_level*2) > centerX)
							zb_offsetX = 0;
						else if(zb_w/(zb_zoom_level*2) + centerX > maxX)
							zb_offsetX = maxX - zb_w/zb_zoom_level;
						else zb_offsetX = centerX - zb_w/(zb_zoom_level*2);
						if(zb_h/(zb_zoom_level*2) > centerY)
							zb_offsetY = 0;
						else if(zb_h/(zb_zoom_level*2) + centerY > maxY)
							zb_offsetY = maxY - zb_h/zb_zoom_level;
						else zb_offsetY = centerY - zb_h/(zb_zoom_level*2);
					}
					show_image(new_surf,offsetX,offsetY,0);
					draw_zoom_box(page_img,((float)scale_width)/page_img->w,*rotation,zb_offsetX,zb_offsetY,offsetX,offsetY,zb_zoom_level);
					pending_action = no_event;
					break;
				}
				float centerY = (offsetX + (new_surf->w > screen->w?screen->w:new_surf->w)/2.0f)/(float)new_surf->w;
				float centerX = (offsetY + (new_surf->h > screen->h?screen->h:new_surf->h)/2.0f)/(float)new_surf->h;
				//SDL_Surface *buffer;
				offsetX_shift = offsetY_shift = 0;
				*rotation+=(poll_command&rotate_cw?1:3);
				*rotation %= 4;
				if(new_surf != page_img) SDL_FreeSurface(new_surf);
				if(full_mode == full_width) {
					if(*rotation & 1)
						scale_width = page_img->w*(screen->w/(float)page_img->h);
					else scale_width = screen->w;
				}
				new_surf = rotozoom(page_img,NULL,scale_width,0,*rotation);
				/*if(!buffer) return init_event(error_ev);
				new_surf = rotate(buffer, *rotation);*/
				if(!new_surf) return init_event(error_ev);
				if(new_surf->w <= screen->w) offsetX = 0;
				else {
					float result;
					if(poll_command & rotate_cw) {
						result = new_surf->w - (centerX * new_surf->w) - screen->w/2.0f;
					} else {
						result = (centerX * new_surf->w) - screen->w/2.0f;
					}
					offsetX = (result > 0?result:0);
					if((offsetX + screen->w) > new_surf->w) offsetX = new_surf->w - screen->w;					
				}
				if(new_surf->h <= screen->h) offsetY = 0;
				else {
					float result;
					if(poll_command & rotate_cw) {
						result = (centerY * new_surf->h) - screen->h/2;
					} else {
						result = new_surf->h - (centerY * new_surf->h) - screen->h/2;
					}
					offsetY = (result > 0?result:0);
					if((offsetY + screen->h) > new_surf->h) offsetY = new_surf->h - screen->h;					
				}
				//if((new_surf != buffer) && (buffer != page_img)) SDL_FreeSurface(buffer);
				if((new_surf->w < screen->w) || (new_surf->h < screen->h))
					clear_screen();
				show_image(new_surf,offsetX,offsetY,1);
			}
			pending_action = no_event;
			break;
			case zoom_in|event_end: case zoom_out|event_end:
			if((pending_action != zoom_in) && (pending_action != zoom_out)) break;
			if((clock() - start)/CLOCKS_PER_SEC < 2) {
				SDL_KillThread(izb_thread); //Yeah yeah, I know.
				izb_thread = NULL;
				if(zbi.zoom_box_is_active) {
					zbi.zoom_box_is_active = 0;
					if((new_surf->w < screen->w) || (new_surf->h < screen->h))
						clear_screen();
					show_image(new_surf,offsetX,offsetY,1);
				} else if(
				  !(
					(poll_command & zoom_in) && (scale_width == (Uint16)(zoom_max*page_img->w))
				  ) && !(
				  	(poll_command & zoom_out) && (scale_width == (Uint16)(zoom_min*page_img->w))
				)) {
					*scale_ratio = ((float)scale_width)/page_img->w;
					float centerX, centerY;
					centerX = (offsetX + (new_surf->w > screen->w?screen->w:new_surf->w)/2.0f)/(float)new_surf->w;
					centerY = (offsetY + (new_surf->h > screen->h?screen->h:new_surf->h)/2.0f)/(float)new_surf->h;
					/*scale_ratio = round(scale_ratio/zoom_amount);
					scale_ratio *= zoom_amount;*/
					*scale_ratio += (poll_command&zoom_in?zoom_amount:-zoom_amount);
					if(*scale_ratio > zoom_max) *scale_ratio = zoom_max;
					else if(*scale_ratio < zoom_min) *scale_ratio = zoom_min;
					if(*scale_ratio == 1) full_mode = full_view;
					scale_width = (Uint16)(*scale_ratio*page_img->w);
					if(new_surf != page_img) SDL_FreeSurface(new_surf);
					new_surf = rotozoom(page_img,NULL,scale_width,0,*rotation);
					if(!new_surf) return init_event(error_ev);
					offsetX_shift = offsetY_shift = 0;
					/*SDL_Surface *buffer = new_surf;
					scale_width = new_surf->w;
					new_surf = rotate(buffer,*rotation);
					if((new_surf != buffer) && (buffer != page_img)) SDL_FreeSurface(buffer);*/
					if(!new_surf) return init_event(error_ev);
					full_mode = custom_zoom;
					if(new_surf->w <= screen->w) offsetX = 0;
					else {
						float result = (centerX * new_surf->w) - screen->w/2;
						offsetX = (result > 0?result:0);
						if((offsetX + screen->w) > new_surf->w) offsetX = new_surf->w - screen->w;
					}
					if(new_surf->h <= screen->h) offsetY = 0;
					else {
						float result = (centerY * new_surf->h) - screen->h/2;
						offsetY = (result > 0?result:0);
						if((offsetY + screen->h) > new_surf->h) offsetY = new_surf->h - screen->h;
					}
					if((new_surf->w < screen->w) || (new_surf->h < screen->h))
						clear_screen();
					show_image(new_surf,offsetX,offsetY,1);
				}
			}
			pending_action = no_event;
			break;
			case zoom_fixed|event_end:
			if(pending_action != zoom_fixed) break;
			if((clock() - start)/CLOCKS_PER_SEC < 2) {
				if(!zbi.zoom_box_is_active) {
					float real_scale_ratio = new_surf->w / (float)page_img->w;
					float centerX = (offsetX + (new_surf->w > screen->w?screen->w:new_surf->w)/2.0f)/(float)new_surf->w;
					float centerY = (offsetY + (new_surf->h > screen->h?screen->h:new_surf->h)/2.0f)/(float)new_surf->h;
					offsetX = offsetY = 0;
					offsetX_shift = offsetY_shift = 0;
					if((*access_int_global(access_fullview_mode) && (full_mode == full_view)) || (full_mode != full_width)) {
						if(new_surf != page_img) SDL_FreeSurface(new_surf);
						*scale_ratio = 0;
						scale_width = screen->w;
						/*new_surf = (*rotation & 1?
							resize(page_img,NULL,0,scale_width):
							resize(page_img,NULL,scale_width,0)
						);*/
						new_surf = (*rotation & 1?
							rotozoom(page_img,NULL,0,scale_width,*rotation):
							rotozoom(page_img,NULL,scale_width,0,*rotation)
						);
						if(!new_surf) return init_event(error_ev);
						scale_width = new_surf->w;
						/*SDL_Surface *buffer = new_surf;
						new_surf = rotate(buffer,*rotation);
						if((new_surf != buffer) && (buffer != page_img)) SDL_FreeSurface(buffer);
						if(!new_surf) return init_event(error_ev);*/
						full_mode = full_width;
						real_scale_ratio = new_surf->w / (float)page_img->w;
						if(new_surf->h <= screen->h) offsetY = 0;
						else {
							float result = (centerY * new_surf->h) - screen->h/2;
							offsetY = (result > 0?result:0);
							if((offsetY + screen->h) > new_surf->h) offsetY = new_surf->h - screen->h;
						}
						if((new_surf->w < screen->w) || (new_surf->h < screen->h))
							clear_screen();
						show_image(new_surf,0,offsetY,1);
					} else {
						if(new_surf != page_img) SDL_FreeSurface(new_surf);
						*scale_ratio = 1;
						scale_width = page_img->w;
						new_surf = rotozoom(page_img,NULL,page_img->w,page_img->h,*rotation);
						full_mode = full_view;
						if(new_surf->w <= screen->w) offsetX = 0;
						else {
							float result = (centerX * new_surf->w) - screen->w/2;
							offsetX = (result > 0?result:0);
							if((offsetX + screen->w) > new_surf->w) offsetX = new_surf->w - screen->w;
						}
						if(new_surf->h <= screen->h) offsetY = 0;
						else {
							float result = (centerY * new_surf->h) - screen->h/2;
							offsetY = (result > 0?result:0);
							if((offsetY + screen->h) > new_surf->h) offsetY = new_surf->h - screen->h;
						}
						if((new_surf->w < screen->w) || (new_surf->h < screen->h))
							clear_screen();
						show_image(new_surf,offsetX,offsetY,1);
					}
				}
			}
			pending_action = no_event;
			break;
			case open_menu|event_end:
			if(pending_action != open_menu) break;
			if((clock() - start)/CLOCKS_PER_SEC < 2) {
				event menu_command = show_menu(book);
				switch(menu_command.command) {
					case close_book: case open_book: 
					case jump_page: case quit_command:
						ret = menu_command;
						quit = menu_command.command;
					break;
					case set_pan_rate:
						pan_amount = *access_int_global(access_pan_rate);
						free_event(menu_command);
					break;
					case set_zb_w: case set_zb_h:
						zbi.zoom_box_is_active = 0;
					default: free_event(menu_command); break;
				}
				if((new_surf->w < screen->w) || (new_surf->h < screen->h))
					clear_screen();
				show_image(new_surf,offsetX,offsetY,0);
				if(zbi.zoom_box_is_active)
					draw_zoom_box(page_img,((float)scale_width)/page_img->w,*rotation,zb_offsetX,zb_offsetY,offsetX,offsetY,zb_zoom_level);
				else flip_screen(1);
			} pending_action = no_event;
			break;
			case pan_event: {
				Sint16 joyX = SDL_JoystickGetAxis(joy,0);
				Sint16 joyY = SDL_JoystickGetAxis(joy,1);
				if(joy_in_use) offsetX_shift = offsetY_shift = 0;
				joy_in_use = 0;
				if(abs(joyX) > dead_zone) {
					offsetX_shift = (char) (joyX*joy_sense);
					joy_in_use = 1;
				}
				if(abs(joyY) > dead_zone) {
					offsetY_shift = (char) (joyY*joy_sense);
					joy_in_use = 1;
				}
			} break;
		}
		if(poll_command & event_end) {
			poll_command ^= event_end;
			if(poll_command & pan_event) {
				delay = 0;
				wait_for_input = 1;
				if(poll_command & pan_horiz) offsetX_shift = 0;
				if(poll_command & pan_vert) offsetY_shift = 0;
			}
		}
		if((offsetX_shift == 0) && (offsetY_shift == 0)) joy_in_use = 0;
		#ifdef PSP
		if(!joy_in_use) {
			if(!SDL_JoystickGetButton(joy,9) && !SDL_JoystickGetButton(joy,7) &&
			  !SDL_JoystickGetButton(joy,8) && !SDL_JoystickGetButton(joy,6)) {
					delay = 0;
					wait_for_input = 1;
			}
			if(!SDL_JoystickGetButton(joy,9) && !SDL_JoystickGetButton(joy,7)) {
					offsetX_shift = 0;
			}
			if(!SDL_JoystickGetButton(joy,8) && !SDL_JoystickGetButton(joy,6)) {
					offsetY_shift = 0;
			}
		}
		#endif
		if(offsetX_shift || offsetY_shift) {
			delay = 50;
			wait_for_input = 0;
			if(!zbi.zoom_box_is_active) {
				Uint16 offsetX_buffer = offsetX, offsetY_buffer = offsetY;
				if((offsetX_shift < 0) && (offsetX < -offsetX_shift))
					offsetX = -offsetX_shift;
				if((offsetY_shift < 0) && (offsetY < -offsetY_shift))
					offsetY = -offsetY_shift;
				if(offsetX_shift) offsetX += offsetX_shift;
				if(offsetY_shift) offsetY += offsetY_shift;
				if(screen->w <= new_surf->w) {
					if (offsetX > new_surf->w - screen->w)
						offsetX = new_surf->w - screen->w;
				}
				else offsetX = 0;
				if(screen->h <= new_surf->h) {
					if (offsetY > new_surf->h - screen->h)
						offsetY = new_surf->h - screen->h;
				}
				else offsetY = 0;
				if((offsetX_buffer != offsetX) || (offsetY_buffer != offsetY)) {
					show_image(new_surf,offsetX,offsetY,1);
				}
			} else {
				if((offsetX_shift < 0) && (zb_offsetX < -offsetX_shift)) {
					if(offsetX > 0) {
						Uint16 buffer = -offsetX_shift - zb_offsetX;
						if(buffer > offsetX) offsetX = 0;
						else offsetX -= buffer;
					}
					zb_offsetX = -offsetX_shift;
				}
				if((offsetY_shift < 0) && (zb_offsetY < -offsetY_shift)) {
					if(offsetY > 0) {
						Uint16 buffer = -offsetY_shift - zb_offsetY;
						if(buffer > offsetY) offsetY = 0;
						else offsetY -= buffer;
					}
					zb_offsetY = -offsetY_shift;
				}
				if(offsetX_shift) zb_offsetX += offsetX_shift;
				if(offsetY_shift) zb_offsetY += offsetY_shift;
				if((*access_int_global(access_zoom_box_w)/zb_zoom_level + zb_offsetX) >= new_surf->w)
					zb_offsetX = new_surf->w - *access_int_global(access_zoom_box_w)/zb_zoom_level;
				if((zb_offsetX + *access_int_global(access_zoom_box_w)/zb_zoom_level) > screen->w) {
					offsetX -= screen->w - zb_offsetX - *access_int_global(access_zoom_box_w)/zb_zoom_level;
					if(offsetX > (new_surf->w - screen->w)) offsetX = new_surf->w - screen->w;
					zb_offsetX = screen->w - *access_int_global(access_zoom_box_w)/zb_zoom_level;
				}
				if((*access_int_global(access_zoom_box_h)/zb_zoom_level + zb_offsetY) >= new_surf->h)
					zb_offsetY = new_surf->h - *access_int_global(access_zoom_box_h)/zb_zoom_level;
				if((zb_offsetY + *access_int_global(access_zoom_box_h)/zb_zoom_level) > screen->h) {
					offsetY -= screen->h - zb_offsetY - *access_int_global(access_zoom_box_h)/zb_zoom_level;
					if(offsetY > (new_surf->h - screen->h)) offsetY = new_surf->h - screen->h;
					zb_offsetY = screen->h - *access_int_global(access_zoom_box_h)/zb_zoom_level;
				}
				if((new_surf->w < screen->w) || (new_surf->h < screen->h))
					clear_screen();
				show_image(new_surf,offsetX,offsetY,0);
				draw_zoom_box(page_img,((float)scale_width)/page_img->w,*rotation,zb_offsetX,zb_offsetY,offsetX,offsetY,zb_zoom_level);
			}
		}
		SDL_Delay(delay);
	}
	if(new_surf != page_img) SDL_FreeSurface(new_surf);
	if(ret.command == no_event) ret.command = quit;
	return ret;
}

inline char* init_block(Uint16 kb) {
	Uint32 bytes = kb*1024;
	char* block = malloc(sizeof(char)*bytes);
	if(!block) return NULL;
	return memset(block,0,bytes);
}
inline event init_event(int ev) {
	event ret;
	ret.command = ev;
	ret.option_text = ret.char_data = NULL;
	ret.int_data = 0;
	return ret;
}
inline char ends_with(const char *reference, const char *against) {
	Uint32 agalen = strlen(against);
	if(agalen > strlen(reference)) return -1;
	reference += strlen(reference) - agalen;
	int i;
	for(i = 0; i < agalen; i++)
		if(tolower(against[i]) != tolower(reference[i]))
			return 0;
	return 1;
}
#ifndef PSP
inline void setXY(int x, int y) {
	int i;
	say("\n");
	for(i = 0; i < x; i++) say(" ");
}
#endif
inline char* make_string(const char *original) {
	char* ret = malloc(sizeof(char)*(strlen(original)+1));
	if(!ret) return NULL;
	strcpy(ret,original);
	ret[strlen(original)] = 0;
	return ret;
}

comic_book open_comic_book(char *dir, char *filename) {
	comic_book book;
	book.current_page = 0;
	book.num_pages = 0;
	book.page_order = NULL;
	book.zip_file = NULL;
	book.type = invalid_comic_type;
	book.localname = make_string(filename);
	char *temp1, *temp2;
	temp1 = temp2 = init_block(1);
	if(!temp1) {
		say("Out of memory");
		next_line();
		return book;
	}
	strncpy(temp1,dir,strlen(dir));
	temp1 += strlen(temp1);
	strncpy(temp1,filename,strlen(filename));
	temp1 = temp2;
	filename = make_string(temp1);
	free(temp1);
	book.filename = filename;
	if((ends_with(filename,".cbz") > 0) || (ends_with(filename,".zip") > 0)) {
		book.type = comic_book_zip;
		book.zip_file = unzOpen(filename);
		if(book.zip_file == NULL) {
			say("Error opening file.");
			next_line();
			unzClose(book.zip_file);
			book.zip_file = NULL;
			return book;
		}
		int status = UNZ_OK;
		do {
			char *block = init_block(1);
			if(!block) {
				say("Out of memory");
				next_line();
				unzClose(book.zip_file);
				book.zip_file = NULL;
				return book;
			}
			unzGetCurrentFileInfo(book.zip_file,NULL,block,1024,NULL,0,NULL,0);
			if(block[1023] != 0) {
				say("Filename in zip too long.");
				next_line();
				unzClose(book.zip_file);
				book.zip_file = NULL;
				free(block);
				return book;
			}
			if((ends_with(block,".jpg") > 0) || (ends_with(block,".jpeg") > 0) ||
			  (ends_with(block,".png") > 0) || (ends_with(block,".tiff") > 0) ||
			  (ends_with(block,".bmp") > 0) || (ends_with(block,".gif") > 0))
				book.num_pages++;
			free(block);
		} while((status = unzGoToNextFile(book.zip_file)) == UNZ_OK);
		if(status != UNZ_END_OF_LIST_OF_FILE) {
			say("Error reading zip.");
			next_line();
			unzClose(book.zip_file);
			book.zip_file = NULL;
			book.num_pages = 0;
			return book;
		} else if(book.num_pages < 1) {
			say("Book is empty.");
			next_line();
			unzClose(book.zip_file);
			book.zip_file = NULL;
			return book;
		}
		book.page_order = malloc(sizeof(char*)*book.num_pages);
		if(!book.page_order) {
			say("Out of memory");
			next_line();
			book.num_pages = 0;
			return book;
		}
		memset(book.page_order,0,book.num_pages*sizeof(char*));
		int i;
		unzGoToFirstFile(book.zip_file);
		status = UNZ_OK;
		for(i = 0; (status == UNZ_OK) && (i < book.num_pages); i++) {
			char *block = init_block(1);
			if(!block) {
				say("Out of memory");
				next_line();
				book.num_pages = 0;
				unzClose(book.zip_file);
				book.zip_file = NULL;
				return book;
			}
			unzGetCurrentFileInfo(book.zip_file,NULL,block,1023,NULL,0,NULL,0);
			if((ends_with(block,".jpg") > 0) || (ends_with(block,".jpeg") > 0) ||
			  (ends_with(block,".png") > 0) || (ends_with(block,".tiff") > 0) ||
			  (ends_with(block,".bmp") > 0) || (ends_with(block,".gif") > 0))
				book.page_order[i] = block;
			else i--;
			status = unzGoToNextFile(book.zip_file);
		}
		unzGoToFirstFile(book.zip_file);
	} else if((ends_with(filename,".cbr") > 0) || (ends_with(filename,".rar") > 0)) {
		book.type = comic_book_rar;
		void *rar_file = rar_open(filename,RAR_OM_LIST);
		if(!rar_file) return book;
		struct RARHeaderData *hd = malloc(sizeof(struct RARHeaderData));
		if(!hd) {
			say("Out of memory");
			next_line();
			return book;
		}
		memset(hd,0,sizeof(struct RARHeaderData));
		int status = 0;
		while(!(status = rar_get_header(rar_file,hd))) {
			char *block = hd->FileName;
			if((ends_with(block,".jpg") > 0) || (ends_with(block,".jpeg") > 0) ||
			  (ends_with(block,".png") > 0) || (ends_with(block,".tiff") > 0) ||
			  (ends_with(block,".bmp") > 0) || (ends_with(block,".gif") > 0))
				book.num_pages++;
			rar_next_file(rar_file);
		}
		rar_close(rar_file);
		if(status != ERAR_END_ARCHIVE) {
			say("Error reading rar.");
			next_line();
			book.num_pages = 0;
			return book;
		} else if(book.num_pages < 1) {
			say("Book is empty.");
			next_line();
			return book;
		}
		rar_file = rar_open(filename,RAR_OM_LIST);
		book.page_order = malloc(sizeof(char*)*book.num_pages);
		if(!book.page_order) {	
			say("Out of memory");
			next_line();
			book.num_pages = 0;
			return book;
		}
		memset(book.page_order,0,book.num_pages*sizeof(char*));
		int i;
		for(i = 0; (i < book.num_pages) && !rar_get_header(rar_file,hd); i++) {
			char *block = hd->FileName;
			if((ends_with(block,".jpg") > 0) || (ends_with(block,".jpeg") > 0) ||
			  (ends_with(block,".png") > 0) || (ends_with(block,".tiff") > 0) ||
			  (ends_with(block,".bmp") > 0) || (ends_with(block,".gif") > 0))
				book.page_order[i] = make_string(block);
			else i--;
			rar_next_file(rar_file);
		}
		rar_close(rar_file);
		free(hd);
	} else return book;
	qsort(book.page_order,book.num_pages,sizeof(char*),cb_compare);
	return book;
}
void close_comic_book(comic_book book) {
	if(book.type == comic_book_zip)
		unzClose(book.zip_file);
	else if(book.type == comic_book_rar) {
	}
	int i;
	for(i = 0; i < book.num_pages; i++) {
		if(book.page_order[i] == NULL) {
			say("Memory leak detected.");
			next_line();
		}
		else free(book.page_order[i]);
	}
	free(book.page_order);
	free(book.filename);
	free(book.localname);
}
comic_book goto_next_page(comic_book book) {
	if(book.current_page >= book.num_pages-1) {
		book.current_page = book.num_pages-1;
	} else book.current_page++;
	return book;
}
comic_book goto_prev_page(comic_book book) {
	if(book.current_page <= 0) {
		book.current_page = 0;
	} else book.current_page--;
	return book;
}
comic_book goto_page(comic_book book, Uint16 page) {
	if(page < book.num_pages) book.current_page = page;
	return book;
}

event show_menu(comic_book *book) {
	char book_is_open = book != NULL;
	char *lwd = make_string(cwd); //Local working directory
	device acc = access_device(access_joystick);
	SDL_Joystick *joy = acc.joystick;
	acc = access_device(access_screen);
	SDL_Surface *screen = acc.screen;
	SDL_Surface *bg = load_background();
	if(bg) show_image(bg,0,0,1);
	else {
		clear_screen();
		flip_screen(1);
	}
	Uint16 current_option = 0, num_options = 0,
		current_submenu = (book_is_open?2:1), supermenu = 0,
		scroll_offset = 0;
	Sint32 minimum = 0, maximum = 0;
	int menu_type;
	static char run_battery_meter = 1;
		//This is static because it is passed to as a pointer to a thread that
		//may remain running after the function exits.
	run_battery_meter = 1; //It needs to be reset every time the function runs
	SDL_Thread *bmt = SDL_CreateThread(battery_meter_thread,&run_battery_meter);
	event action, *options = NULL;
	action = init_event(no_event);
	while(action.command == no_event) {
		current_option = 0;
		switch(current_submenu) {
			case 1: { //Open book menu
				supermenu = (book_is_open?2:0);
				menu_type = option_menu;
				Uint16 num_books;
				event *options_buf = NULL;
				num_books = file_list(lwd,&options_buf,1);
				num_options = num_books?num_books+2:3;
				options = malloc(sizeof(event)*num_options);
				if(!options) {
					run_battery_meter = 0;
					setXY(0,0);
					free(lwd);
					say("Out of memory");
					next_line();
					return init_event(error_ev);
				}
				if(num_books) {
					memcpy(options,options_buf,num_books*sizeof(event));
					free(options_buf);
				} else {
					options[0].option_text = make_string("Please place comic books in the MS_ROOT/comics folder");
					options[0].char_data = NULL;
					options[0].int_data = 0;
					options[0].command = no_event;

					#ifdef PSP
					sceIoMkdir("ms0:/comics",0777);
					#endif
				}
				options[num_options-2].option_text = make_string("Reload menu");
				options[num_options-2].char_data = NULL;
				options[num_options-2].int_data = 0;
				options[num_options-2].command = open_menu;
			} break;
			case 2: { //In-book menu
				supermenu = 0;
				menu_type = option_menu;
				num_options = 12;
				options = malloc(sizeof(event)*num_options);
				if(!options) {
					run_battery_meter = 0;
					setXY(0,0);
					free(lwd);
					say("Out of memory");
					next_line();
					return init_event(error_ev);
				}
				options[0].option_text = make_string("Open comic book...");
				options[0].command = open_submenu;
				options[0].int_data = 1;
				options[0].char_data = NULL;

				options[1].option_text = make_string("Close comic book");
				options[1].command = close_book;
				options[1].int_data = 0;
				options[1].char_data = NULL;

				options[2].option_text = make_string("Next comic book");
				options[2].command = aux_command;
				options[2].int_data = next_book;
				options[2].char_data = NULL;

				options[3].option_text = make_string("Previous comic book");
				options[3].command = aux_command;
				options[3].int_data = prev_book;
				options[3].char_data = NULL;

				options[4].option_text = make_string("Jump to page...");
				options[4].command = open_submenu;
				options[4].int_data = 3;
				options[4].char_data = NULL;

				options[5].option_text = make_string("Load bookmark");
				options[5].command = aux_command;
				options[5].int_data = retrieve_bookmark;
				options[5].char_data = NULL;

				options[6].option_text = make_string("Set to bookmark");
				options[6].command = aux_command;
				options[6].int_data = add_bookmark;
				options[6].char_data = NULL;

				options[7].option_text = make_string("More bookmark operations...");
				options[7].command = open_submenu;
				options[7].int_data = 8;
				options[7].char_data = NULL;

				options[8].option_text = make_string("Configuration...");
				options[8].command = open_submenu;
				options[8].int_data = 5;
				options[8].char_data = NULL;

				options[9].option_text = make_string("About");
				options[9].command = popup;
				options[9].int_data = 0;
				options[9].char_data = NULL;

				options[10].option_text = make_string("Quit");
				options[10].command = quit_command;
				options[10].int_data = 0;
				options[10].char_data = NULL;
			} break;
			case 3: { //Jump to page menu
				supermenu = 2;
				menu_type = option_menu;
				num_options = book->num_pages + 1;
				options = malloc(sizeof(event)*num_options);
				if(!options) {
					run_battery_meter = 0;
					setXY(0,0);
					free(lwd);
					say("Out of memory");
					next_line();
					return init_event(error_ev);
				}

				int i;
				char *buffer, *buffer2;
				for(i = 0; i < book->num_pages; i++) {
					buffer = malloc(sizeof(char)*80);
					memset(buffer,0,sizeof(char)*80);
					sprintf(buffer,"Page %u: ",i+1);
					buffer2 = strrchr(book->page_order[i],'/');
					if(!buffer2) buffer2 = strrchr(book->page_order[i],'\\');
					strncat(buffer,(buffer2?buffer2+1:book->page_order[i]),80-strlen(buffer));
					options[i].option_text = make_string(ellipsis(buffer,0));
					free(buffer);
					options[i].command = jump_page;
					options[i].int_data = i;
					options[i].char_data = NULL;
				}
			} break;
			case 4: { //Adjust clock frequency
				supermenu = 5;
				menu_type = number_menu;
				num_options = 1;
				options = malloc(sizeof(event));
				if(!options) {
					run_battery_meter = 0;
					setXY(0,0);
					free(lwd);
					say("Out of memory");
					next_line();
					return init_event(error_ev);
				}
				options->option_text = make_string("Set clock frequency (MHz)");
				options->command = clock_adjust;
				options->int_data =
				#ifdef PSP
				scePowerGetCpuClockFrequency();
				#else
				222;
				#endif
				options->char_data = NULL;
				maximum = 333;
				minimum = 20;
			} break;
			case 5: { //Config menu
				supermenu = 2;
				menu_type = option_menu;
				num_options = 15;
				options = malloc(sizeof(event)*num_options);
				if(!options) {
					run_battery_meter = 0;
					setXY(0,0);
					free(lwd);
					say("Out of memory");
					next_line();
					return init_event(error_ev);
				}
				options[0].option_text = (*access_int_global(access_resize)?
					make_string("Change resize method to nearest neighbor (faster, blockier)"):
					make_string("Change resize method to resample (slower, smoother)")
				);
				options[0].command = change_resize_method;
				options[0].int_data = -1;
				options[0].char_data = NULL;

				options[1].option_text = make_string("Adjust clock frequency...");
				options[1].command = open_submenu;
				options[1].int_data = 4;
				options[1].char_data = NULL;

				options[2].option_text = make_string("Adjust pan rate...");
				options[2].command = open_submenu;
				options[2].int_data = 6;
				options[2].char_data = NULL;

				options[3].option_text = (*access_int_global(access_zoom_persist)?
					make_string("Turn off zoom level persistence"):
					make_string("Turn on zoom level persistence")
				);
				options[3].command = toggle_global;
				options[3].int_data = access_zoom_persist;
				options[3].char_data = NULL;

				options[4].option_text = (*access_int_global(access_rotate_persist)?
					make_string("Turn off rotation persistence"):
					make_string("Turn on rotation persistence")
				);
				options[4].command = toggle_global;
				options[4].int_data = access_rotate_persist;
				options[4].char_data = NULL;

				options[5].option_text = make_string("Adjust menu scroll skip rate");
				options[5].command = open_submenu;
				options[5].int_data = 7;
				options[5].char_data = NULL;

				options[6].option_text = (*access_int_global(access_manga_mode)?
					make_string("Turn off manga mode"):
					make_string("Turn on manga mode")
				);
				options[6].command = toggle_global;
				options[6].int_data = access_manga_mode;
				options[6].char_data = NULL;

				options[7].option_text = (*access_int_global(access_bookmark_on_load)?
					make_string("Do not jump to bookmark upon loading comic book"):
					make_string("Jump to bookmark upon loading comic book")
				);
				options[7].command = toggle_global;
				options[7].int_data = access_bookmark_on_load;
				options[7].char_data = NULL;

				options[8].option_text = (*access_int_global(access_fullview_mode)?
					make_string("Turn off full view mode"):
					make_string("Turn on full view mode")
				);
				options[8].command = toggle_global;
				options[8].int_data = access_fullview_mode;
				options[8].char_data = NULL;

				options[9].option_text = make_string("Set zoom box width");
				options[9].command = open_submenu;
				options[9].int_data = 9;
				options[9].char_data = NULL;

				options[10].option_text = make_string("Set zoom box height");
				options[10].command = open_submenu;
				options[10].int_data = 10;
				options[10].char_data = NULL;

				options[11].option_text = (*access_int_global(access_bookmark_on_load)?
					make_string("Turn off precaching"):
					make_string("Turn on precaching (may be buggy)")
				);
				options[11].command = toggle_global;
				options[11].int_data = access_precaching;
				options[11].char_data = NULL;

				options[12].option_text = make_string("Save configuration");
				options[12].command = aux_command;
				options[12].int_data = save_config_command;
				options[12].char_data = NULL;

				options[13].option_text = make_string("Load configuration");
				options[13].command = aux_command;
				options[13].int_data = load_config_command;
				options[13].char_data = NULL;
			} break;
			case 6: { //Adjust pan rate menu
				supermenu = 5;
				menu_type = number_menu;
				maximum = pan_max;
				minimum = pan_min;
				options = malloc(sizeof(event));
				num_options = 1;

				options->option_text = make_string("Adjust pan rate:");
				options->command = set_pan_rate;
				options->int_data = *access_int_global(access_pan_rate);
				options->char_data = NULL;
			} break;
			case 7: { //Adjust scroll skip amount
				supermenu = 5;
				menu_type = number_menu;
				maximum = 30;
				minimum = 1;
				options = malloc(sizeof(event));
				num_options = 1;

				options->option_text = make_string("Set scroll skip:");
				options->command = set_scroll_skip;
				options->int_data = *access_int_global(access_scroll_skip);
				options->char_data = NULL;
			} break;
			case 8: { //Bookmark menu
				supermenu = 2;
				menu_type = option_menu;
				num_options = 4;
				options = malloc(sizeof(event)*num_options);
				if(!options) {
					run_battery_meter = 0;
					setXY(0,0);
					free(lwd);
					say("Out of memory");
					next_line();
					return init_event(error_ev);
				}

				options[0].option_text = make_string("Delete bookmark for this comic book");
				options[0].command = aux_command;
				options[0].int_data = delete_bookmarks_command;
				options[0].char_data = NULL;

				options[1].option_text = make_string("Delete all bookmarks");
				options[1].command = aux_command;
				options[1].int_data = delete_all_bookmarks_command;
				options[1].char_data = NULL;

				options[2].option_text = make_string("Purge bookmarks");
				options[2].command = aux_command;
				options[2].int_data = purge_bookmarks_command;
				options[2].char_data = NULL;
			} break;
			case 9: { //Set zoom box width
				supermenu = 5;
				menu_type = number_menu;
				maximum = 180;
				minimum = 20;
				options = malloc(sizeof(event));
				num_options = 1;

				options->option_text = make_string("Set zoom box width:");
				options->command = set_zb_w;
				options->int_data = *access_int_global(access_zoom_box_w);
				options->char_data = NULL;
			} break;
			case 10: { //Set zoom box height
				supermenu = 5;
				menu_type = number_menu;
				maximum = 250;
				minimum = 20;
				options = malloc(sizeof(event));
				num_options = 1;

				options->option_text = make_string("Set zoom box height:");
				options->command = set_zb_h;
				options->int_data = *access_int_global(access_zoom_box_h);
				options->char_data = NULL;
			} break;
			default:
				say("Internal error: submenu does not exist");
				next_line();
				return init_event(error_ev);
			break;
		}
		if(menu_type == option_menu) {
			event *last_option = options + num_options - 1;
			last_option->option_text = make_string("< Go back");
			last_option->char_data = NULL;
			if(supermenu) {
				last_option->command = open_submenu;
				last_option->int_data = supermenu;
			} else {
				last_option->command = event_end;
				last_option->int_data = 0;
			}
		}
		//Menu display/interaction
		int poll_command;
		char redraw_menu = 1, scrolling = 0, wait_for_input = 1;
		switch(menu_type) {
			case option_menu:
			while(action.command == no_event) {
				if(redraw_menu) {
					#ifdef PSP
					pspDebugScreenSetOffset((int)screen->pixels & 0x1FFFFF);
					#endif
					run_battery_meter = 2;
					if(bg) show_image(bg,0,0,0);
					else clear_screen();
					battery_meter();
					setXY(0,0);
					if(scroll_offset > 0) say("-- More --");
					setXY(0,1);
					Uint16 i, i_max = (options_per_page<(num_options-scroll_offset)
						?options_per_page:num_options-scroll_offset);
					for(i = 0; i < i_max; i++) {
						setXY(0,i+1);
						int j = i + scroll_offset;
						#ifdef PSP
						if(j == current_option) pspDebugScreenSetBackColor(selected_color);
						#else
						if(j == current_option) say("* ");
						#endif
						say("%s",options[j].option_text);
						#ifdef PSP
						pspDebugScreenSetBackColor(0);
						#endif
					}
					if((num_options-scroll_offset)>options_per_page) {
						setXY(0,i+1);
						say("-- More --");
					}
					if(book) {
						setXY(0,33);
						char *status = init_block(1);
						if(strlen(book->localname) > 50) {
							strncpy(status,book->localname,47);
							status[47] = '.';
							status[48] = '.';
							status[49] = '.';
						} else strcpy(status,book->localname);
						char *page_info = init_block(1);
						sprintf(page_info,": %u/%u",book->current_page + 1,book->num_pages);
						strcat(status,page_info);
						free(page_info);
						say_left(status);
						free(status);
					}
					redraw_menu = 0;
					run_battery_meter = 1;
					flip_screen(0);
				}
				poll_command = handle_input(wait_for_input);
				switch(poll_command) {
					case select_option|event_end: action = options[current_option]; break;
					case go_back|event_end: {
						if(current_submenu == 1) {
							free(lwd);
							lwd = make_string(cwd);
						}
						if(supermenu) {
							action.command = open_submenu;
							action.int_data = supermenu;
						} else action.command = event_end;
					} break;
					case pan_down: {
						scrolling = 1;
						wait_for_input = 0;
					} break;
					case pan_up: {
						scrolling = -1;
						wait_for_input = 0;
					} break;
					case next_page: {
						scrolling = *access_int_global(access_scroll_skip);
						wait_for_input = 0;
					} break;
					case prev_page: {
						scrolling = -*access_int_global(access_scroll_skip);;
						wait_for_input = 0;
					} break;
					case open_menu|event_end: if(book) action.command = event_end; break;
				}
				if(scrolling) {
					#ifdef PSP
					if(!SDL_JoystickGetButton(joy,8) &&
					  !SDL_JoystickGetButton(joy,6) &&
					  !SDL_JoystickGetButton(joy,4) &&
					  !SDL_JoystickGetButton(joy,5)
					) {
						scrolling = 0;
						wait_for_input = 1;
					}
					else {
					#endif
						current_option = (current_option + num_options + scrolling) % num_options;
						redraw_menu = 1;
						//Adjust menu position
						if(num_options > options_per_page) {
							if(scrolling > 0) {
								if((current_option + 1) > (scroll_offset + options_per_page))
									scroll_offset = (1 + current_option) - options_per_page;
								else if(((Sint32)current_option - scrolling) < 0) scroll_offset = 0;
							} else {
								if(current_option < scroll_offset) scroll_offset = current_option;
								else if(((Uint32)current_option - scrolling) >= num_options)
									scroll_offset = num_options - options_per_page;
							}
						}
					#ifndef PSP
						scrolling = 0;
						wait_for_input = 1;
					#else
					}
					#endif
				}
				if(scrolling) SDL_Delay(menu_scroll_rate);
			} break;
			case number_menu: {
				Uint8 num_digits0, num_digits1, num_digits;
				Sint32 max_buf = maximum, min_buf = minimum, value = options->int_data;
				for(num_digits0 = 0; max_buf; max_buf /= 10) num_digits0++;
				for(num_digits1 = 0; min_buf; min_buf /= 10) num_digits1++;
				num_digits = (num_digits0 > num_digits1?num_digits0:num_digits1);
				while(action.command == no_event) {
					poll_command = handle_input(wait_for_input);
					switch(poll_command) {
						case select_option|event_end: {
							if(current_option == 0) {
								action = *options;
								action.int_data = value;
							} else if(current_option == 1) {
								if(supermenu) {
									action.command = open_submenu;
									action.int_data = supermenu;
								} else action.command = event_end;
							} else {
								current_option = 0;
								redraw_menu = 1;
							}
						} break;
						case go_back|event_end: {
							if(supermenu) {
								action.command = open_submenu;
								action.int_data = supermenu;
							} else action.command = event_end;
						}break;
						case open_menu|event_end: if(book) action.command = event_end; break;
						case pan_down: {
							scrolling = 1;
							wait_for_input = 0;
						} break;
						case pan_up: {
							scrolling = -1;
							wait_for_input = 0;
						} break;
						case pan_right: {
							current_option = (current_option+1)%(3+num_digits);
							redraw_menu = 1;
						} break;
						case pan_left: {
							current_option = (current_option+2+num_digits)%(3+num_digits);
							redraw_menu = 1;
						} break;
					}
					if(scrolling) {
						#ifdef PSP
						if(!SDL_JoystickGetButton(joy,8) && !SDL_JoystickGetButton(joy,6)) {
							scrolling = 0;
							wait_for_input = 1;
						}
						else {
						#endif
							if(scrolling > 0) {
								if(current_option >= 3) {
									value -= (Sint32)pow(10,num_digits-current_option+2);
								}
							} else {
								if(current_option >= 3) {
									value += (Sint32)pow(10,num_digits-current_option+2);
								}
							}
							redraw_menu = 1;
							if(value > maximum) value = maximum;
							if(value < minimum) value = minimum;
						#ifndef PSP
							scrolling = 0;
							wait_for_input = 1;
						#else
						}
						#endif
					}
					if(redraw_menu) {
						run_battery_meter = 2;
						if(bg) show_image(bg,0,0,1);
						else {
							clear_screen();
							flip_screen(1);
						}
						battery_meter();
						setXY(0,0);
						if(options->option_text) say("%s",options->option_text);
						setXY(0,1);
						#ifdef PSP
						if(current_option == 2) pspDebugScreenSetBackColor(selected_color);
						#else
						if(current_option == 2) say("* ");
						#endif
						if(value < 0) say("-");
						else say("+");
						#ifdef PSP
						pspDebugScreenSetBackColor(0);
						#endif
						int i;
						for(i = num_digits-1; i >= 0; i--) {
							#ifdef PSP
							if(current_option == num_digits-i+2) pspDebugScreenSetBackColor(selected_color);
							#else
							if(current_option == num_digits-i+2) say("* ");
							#endif
							char digit = abs((Sint32)(value/pow(10,i))%10);
							say("%u",digit);
							#ifdef PSP
							pspDebugScreenSetBackColor(0);
							#endif
						}
						setXY(0,2);
						#ifdef PSP
						if(current_option == 0) pspDebugScreenSetBackColor(selected_color);
						#else
						if(current_option == 0) say("* ");
						#endif
						say("Accept");
						#ifdef PSP
						pspDebugScreenSetBackColor(0);
						#endif
						setXY(10,2);
						#ifdef PSP
						if(current_option == 1) pspDebugScreenSetBackColor(selected_color);
						#else
						if(current_option == 1) say("* ");
						#endif
						say("Cancel");
						#ifdef PSP
						pspDebugScreenSetBackColor(0);
						#endif
						redraw_menu = 0;
						run_battery_meter = 2;
					}
					if(scrolling) SDL_Delay(menu_scroll_rate);
				}
			} break;
			default:
				//TODO give error
			break;
		}
		if(bg) show_image(bg,0,0,1);
		else {
			clear_screen();
			flip_screen(1);
		}
		switch(action.command) {
			case go_back:
				if(supermenu == 0) {
					action.command = event_end;
					break;
				}
			case open_submenu:
				if(current_submenu == 1) {
					free(lwd);
					lwd = make_string(cwd);
				}
				current_submenu = action.int_data;
			case open_menu:
				free_event(action);
				action = init_event(no_event);
				current_option = 0;
				scroll_offset = 0;
			break;
			case popup:
				run_battery_meter = 2;
				show_popup(action.int_data);
				run_battery_meter = 1;
				free_event(action);
				action = init_event(no_event);
			break;
			case open_book:
				if(action.char_data[strlen(action.char_data)-1] == '/') {
					char *buffer;
					if(strcmp(action.char_data,"../") != 0) {
						buffer = malloc(sizeof(char)*
							(strlen(action.char_data)+strlen(lwd)+1));
						memset(buffer,0,strlen(action.char_data)+strlen(lwd)+1);
						strcpy(buffer,lwd);
						strcat(buffer,action.char_data);
					} else {
						lwd[strlen(lwd)-1] = 0;
						char *temp = strrchr(lwd,'/');
						temp[1] = 0;
						buffer = make_string(lwd);
					}
					free(lwd);
					lwd = buffer;
					//free_event(action);
					action = init_event(no_event);
					current_option = 0;
				} else {
					free(cwd);
					cwd = lwd;
				}
			break;
		}
		int i;
		for(i = 0; i < num_options; i++) {
			if(
			  (options[i].option_text != action.option_text) &&
			  (options[i].char_data != action.char_data)
			) free_event(options[i]);
		}
		free(options);
	}
	if(cwd != lwd) free(lwd);
	switch(action.command) {
		case clock_adjust:
			set_clock(action.int_data);
		break;
		case change_resize_method: {
			int *method = access_int_global(access_resize);
			if(action.int_data == -1)
				*method = (*method == resize_nn?resample:resize_nn);
		} break;
		case toggle_global: {
			int *global = access_int_global(action.int_data);
			if(global) *global = !*global;
		} break;
		case set_pan_rate: {
			*access_int_global(access_pan_rate) = action.int_data;
		} break;
		case set_scroll_skip: {
			*access_int_global(access_scroll_skip) = action.int_data;
		} break;
		case set_zb_w: {
			*access_int_global(access_zoom_box_w) = action.int_data;
		} break;
		case set_zb_h: {
			*access_int_global(access_zoom_box_h) = action.int_data;
		} break;
		case aux_command:
		switch(action.int_data) {
			case next_book: {
				if(book) {
					action.command = open_book;
					action.char_data = next_file_name(cwd,book->localname);
				} else {
					free_event(action);
					action = init_event(event_end);
				}
			} break;
			case prev_book: {
				if(book) {
					action.command = open_book;
					action.char_data = prev_file_name(cwd,book->localname);
				} else {
					free_event(action);
					action = init_event(event_end);
				}
			}
			break;
			case load_config_command: load_config(); break;
			case save_config_command: save_config(); break;
			case add_bookmark: save_bookmark(book); break;
			case retrieve_bookmark: {
				Sint32 page = get_bookmark(book->filename);
				free_event(action);
				if(page >= 0) {
					action = init_event(jump_page);
					action.int_data = page;
				} else action = init_event(no_event);
			} break;
			case delete_bookmarks_command: delete_bookmarks(book->filename);
			break;
			case delete_all_bookmarks_command: delete_all_bookmarks(); break;
			case purge_bookmarks_command: purge_bookmarks(); break;
		} break;
	}
	run_battery_meter = 0;
	return action;
}
inline void free_event(event mo) {
	if(mo.char_data) free(mo.char_data);
	if(mo.option_text && (mo.char_data != mo.option_text)) free(mo.option_text);
}
Uint16 file_list(char *dir, event **ptr_to_list, char get_dirs) {
	Uint16 num_books = 0;
	event *options = NULL;
	*ptr_to_list = NULL;
	DIR *dirh = opendir(dir);
	if(!dirh) return 0;
	struct dirent *filent;
	while((filent = readdir(dirh))) {
		size_t namelen;
		#ifdef PSP
		namelen = strlen(filent->d_name);
		#else
		namelen = filent->d_namlen;
		#endif
		char *temp = malloc(sizeof(char)*(namelen+1));
		temp[namelen] = 0;
		if(namelen+strlen(dir) > 250) continue;
		strncpy(temp,filent->d_name,namelen);
		if(get_dirs && dirent_is_dir(filent,dir)) {
			if(strcmp(temp,".") != 0) num_books++;
		} else if((
		  (ends_with(temp,".cbz") == 1) ||
		  (ends_with(temp,".zip") == 1) ||
		  (ends_with(temp,".cbr") == 1) ||
		  (ends_with(temp,".rar") == 1)) &&
		(get_dirs || !dirent_is_dir(filent,dir))) num_books++;
		free(temp);
	}
	closedir(dirh);
	if(num_books) options = malloc(sizeof(event)*num_books);
	else return 0;
	if(options == NULL) return 0;
	dirh = opendir(dir);
	int i;
	for(i = 0; (filent = readdir(dirh)); i++) {
		size_t namelen;
		#ifdef PSP
		namelen = strlen(filent->d_name);
		#else
		namelen = filent->d_namlen;
		#endif
		char *temp = malloc(sizeof(char)*(namelen+1));
		temp[namelen] = 0;
		strncpy(temp,filent->d_name,namelen);
		if(namelen+strlen(dir) <= 250) {
			if(get_dirs && dirent_is_dir(filent,dir)) {
				if(strcmp(temp,".") == 0) {
					free(temp);
					i--;
					continue;
				}
				char *buffer = malloc(sizeof(char)*(namelen+2));
				strcpy(buffer,temp);
				free(temp);
				buffer[namelen] = '/';
				namelen++;
				temp = buffer;
			}
			else if(!(
			  (ends_with(temp,".cbz") == 1) ||
			  (ends_with(temp,".zip") == 1) ||
			  (ends_with(temp,".cbr") == 1) ||
			  (ends_with(temp,".rar") == 1)) ||
			(!get_dirs && dirent_is_dir(filent,dir))) {
				free(temp);
				i--;
				continue;
			}
			if(namelen > 67) {
				options[i].char_data = make_string(temp);
				temp[64] = temp[65] = temp[66] = '.';
				options[i].option_text = temp;
				temp[67] = 0;
				options[i].command = open_book;
			} else {
				options[i].char_data = temp;
				options[i].char_data[namelen] = 0;
				options[i].command = open_book;
				options[i].option_text = options[i].char_data;
			}
		} 
	}
	closedir(dirh);
	qsort(options,num_books,sizeof(event),event_compare);
	*ptr_to_list = options;
	return num_books;
}
char dirent_is_dir(struct dirent *filent, char *dir) {
	#ifdef PSP
	return filent->d_stat.st_attr & FIO_SO_IFDIR;
	#else
	size_t namlen = strlen(dir)+filent->d_namlen+1;
	char *temp = malloc(sizeof(char)*namlen);
	memset(temp,0,namlen);
	strcpy(temp,dir);
	strncat(temp,filent->d_name,filent->d_namlen);
	DIR *dirh = opendir(temp);
	free(temp);
	if(dirh) {
		closedir(dirh);
		return 1;
	} else return 0;
	#endif
}
void battery_meter() {
	#ifdef PSP
	setXY(61,0);
	if(scePowerIsPowerOnline())
		pspDebugScreenSetTextColor(0x00FFFF);
	else if(scePowerIsLowBattery())
		pspDebugScreenSetTextColor(0x0000FF);
	say("[%03u%%]",scePowerGetBatteryLifePercent());
	pspDebugScreenSetTextColor(0xFFFFFF);
	#endif
}
int battery_meter_thread(void *is_running) {
	char *running = (char*)is_running;
	while(*running) {
		while(*running == 2) SDL_Delay(100);
		battery_meter();
		SDL_Delay(10000);
	}
	return 0;
}

char show_popup(int which) {
	char *filename;
	switch(which) {
		case 0: filename = "about.png"; break; //About screen
		default: return 0;
	}
	SDL_Surface *img = IMG_Load(filename);
	if(!img) {
		say("SDL_image error: %s",IMG_GetError());
		next_line();
		return 0;
	}
	char temp = show_image(img,0,0,1);
	SDL_FreeSurface(img);
	int input;
	clock_t start = 0;
	int action = no_event;
	while((input = handle_input(1)) != error_ev) {
		if(!(input & pan_event) && (input != no_event)) {
			if(input & event_end) {
				if((clock()-start)/CLOCKS_PER_SEC < 2) break;
				else start = action = 0;
			}
			else if(action == no_event) {
				start = clock();
				action = input;
			}
		}
	}
	return temp;
}

char* extract_file(comic_book *book, Uint16 page, size_t *size) {
	if(page > book->num_pages) return NULL;
	if(book->type == comic_book_rar) {
		char *buffer;
		if(rar_get_file_by_name(book->filename,book->page_order[page],&buffer,size)) {
			return NULL;
		}
		return buffer;
	} else if(book->type == comic_book_zip) {
		if(unzLocateFile(book->zip_file,book->page_order[page],1)
		  != UNZ_OK) {
			  return NULL;
		}
		return extract_file_from_zip(book->zip_file, size);
	} else {
		say("Internal error handling archive");
		next_line();
		return NULL;
	}
}
char* extract_file_from_zip(unzFile zip, size_t *size) {
	unz_file_info file_info;
	unzGetCurrentFileInfo(zip,&file_info,NULL,0,NULL,0,NULL,0);
	unzOpenCurrentFile(zip);
	char *block = malloc(sizeof(char)*file_info.uncompressed_size);
	if(!block) return NULL;
	memset(block,0,sizeof(char)*file_info.uncompressed_size);
	int output = unzReadCurrentFile(zip, (voidp)block,file_info.uncompressed_size);
	if(output < 0) {
		say("Error reading zip.");
		next_line();
		unzCloseCurrentFile(zip);
		free(block);
		return NULL;
	}
	if(size) *size = output;
	if(!unzeof(zip)) {
		say("File too big.");
		next_line();
		unzCloseCurrentFile(zip);
		free(block);
		return NULL;
	}
	unzCloseCurrentFile(zip);
	return block;
}

char* next_file_name(char *dir, char *cur_file) {
	event *list = NULL;
	char *ret = NULL;
	Uint16 num_books = file_list(dir,&list,0);
	int i, j = -1;
	for(i = 0; i < num_books; i++) {
		if((j == -1) && (cb_compare(&cur_file,&(list[i].char_data)) < 0)) {
			j = i;
			ret = make_string(list[i].char_data);
		}
		if(i) free_event(list[i]);
	}
	if(j == -1) ret = make_string(list[0].char_data);
	free_event(list[0]);
	free(list);

	return ret;
}
char* prev_file_name(char *dir, char *cur_file) {
	event *list = NULL;
	char *ret = NULL;
	Uint16 num_books = file_list(dir,&list,0);
	int i, j = -1;
	for(i = num_books - 1; i >= 0; i--) {
		if((j == -1) && (cb_compare(&cur_file,&(list[i].char_data)) > 0)) {
			j = i;
			ret = make_string(list[i].char_data);
		}
		if(i != num_books - 1) free_event(list[i]);
	}
	if(j == -1) ret = make_string(list[num_books - 1].char_data);
	free_event(list[num_books - 1]);
	free(list);

	return ret;
}

inline int cb_compare(const void* foo, const void* bar) {
	int result = strncasecmp(*(char**)foo,*(char**)bar,1024);
	return result;
}
inline int event_compare(const void* foo, const void* bar) {
	int result = strncasecmp(((event*)foo)->char_data,((event*)bar)->char_data,1024);
	return result;
}
inline void next_line() {
	#ifdef PSP
	pspDebugScreenSetXY(0,pspDebugScreenGetY()+1);
	SDL_Delay(2500);
	#else
	say("\n");
	fflush(0);
	#endif
}
inline void say_left(const char *string) {
	setXY(
		67 - strlen(string),
		#ifdef PSP
		pspDebugScreenGetY()
		#else
		0
		#endif
	);
	say("%s",string);
}
inline char* ellipsis(char *string, Uint16 offset) {
	if(offset >= 67) return NULL;
	if((strlen(string) + offset) > 67) {
		string[66-offset] = string[65-offset] = string[64-offset] = '.';
		string[67] = 0;
	}
	return string;
}

int main(int argc, char* argv[]) {
	#ifdef PSP
	initial_clock = scePowerGetCpuClockFrequency();
	atexit(_atexit);
	#endif
	cwd = make_string(cb_dir);
	if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_JOYSTICK) < 0) {
		sdl_error();
		return 1;
	}
	init_video(480,272);
	event obtained_command;
	obtained_command.command = no_event;
	obtained_command.option_text = obtained_command.char_data = NULL;
	obtained_command.int_data = 0;
	load_config();
	load_bookmarks();
	while(obtained_command.command != quit_command) {
		free_event(obtained_command);
		obtained_command = show_menu(NULL);
		while(!(obtained_command.command & (close_book|quit_command|error_ev|open_menu))) {
			switch(obtained_command.command) {
				case open_book: {
					char *temp = make_string(obtained_command.char_data);
					free_event(obtained_command);
					obtained_command = show_book(cwd,temp);
					free(temp);
				} break;
				default:
				obtained_command.command = quit_command;
				break;
			}
		}
	}
	show_image(load_background(),0,0,1);
	setXY(0,0);
	say("Closing program...");

	return 0;
}

#ifdef __cplusplus
}
#endif
