/**
    PSPComic - pspcomic.h
    Copyright (C) 2007  Jeffrey P.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
**/
#ifndef PSPComic_h
#define PSPComic_h

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <dirent.h>
#include <ctype.h>
#include <time.h>
#include <malloc.h>

#include "SDL.h"
#include "SDL_thread.h"
#include <SDL_image/SDL_image.h>
#include <SDL/SDL_gfxPrimitives.h>
#include <SDL/SDL_rotozoom.h>

#include "zip/unzip.h"
#include "rar.hpp"
#ifdef __cplusplus
}
#include "tinyxml/tinyxml.h"
extern "C" {
#endif
#ifdef PSP
const float joy_sense = 0.0012f;
#else
#ifdef fflush
#undef fflush
#endif
const float joy_sense = 0.001f;
#endif

//const Uint8 pan_delay = 20;
const Uint8 pan_delay = 0;
const Uint16 dead_zone = 6000;

const Uint16 menu_scroll_rate = 175;

const float zoom_amount = 0.1f;
const float zoom_max = 1.0f;
const float zoom_min = 0.25f;
const Uint8 pan_max = 99;
const Uint8 pan_min = 5;
const Uint8 options_per_page = 32;

#ifdef PSP
#include <pspkernel.h>
#include <pspdisplay.h>
#include <pspdebug.h>
#include <pspctrl.h>
#include <psppower.h>
#include <pspgu.h>
#include <pspgum.h>
#ifdef EXPERIMENTAL
#include <malloc.h>
#endif
#define say pspDebugScreenPrintf
#define setXY pspDebugScreenSetXY
#define cb_dir "ms0:/comics/"
#else
#define say printf
#ifdef unix
#define cb_dir "~/comics/"
#else
#define cb_dir "C:/comics/"
#endif
#endif

enum {
	no_event = 0, event_end = 1,
	pan_up = 2, pan_right = 4, pan_down = 8, pan_left = 16,
	rotate_cw = 32, rotate_ccw = 64,
	zoom_in = 128, zoom_out = 256, zoom_fixed = 512,
	go_back = 128, select_option = 256,
	next_page = 1024, prev_page = 2048,
	open_menu = 4096, quit_command = 8192, open_submenu = 16384,
	open_book = 32768, close_book = 65536, jump_page = 131072,
	change_resize_method = 262144, clock_adjust = 524288,
	popup = 1048576, toggle_global = 2097152, set_pan_rate = 4194304,
	set_scroll_skip = 8388608, set_zb_w = 167772016, set_zb_h = 335544032,
	aux_command = 536870912, error_ev = 1073741824
};
#define pan_event (pan_up|pan_right|pan_down|pan_left)
#define pan_horiz (pan_left|pan_right)
#define pan_vert (pan_up|pan_down)
const Uint32 selected_color = 0x909090;
enum { //Auxiliary commands
	load_config_command, save_config_command,
	retrieve_bookmark, add_bookmark,
	next_book, prev_book,
	delete_bookmarks_command, delete_all_bookmarks_command,
	purge_bookmarks_command
};
enum { custom_zoom, full_width, full_view }; //Full dimension
enum { invalid_comic_type = 0, comic_book_zip, comic_book_rar }; //Comic book file type
enum { access_screen, access_joystick }; //Access device codes
enum { //Access globals
	access_resize = 1,
	access_pan_rate, access_zoom_persist, access_rotate_persist,
	access_scroll_skip, access_precaching, access_manga_mode,
	access_zoom_box_w, access_zoom_box_h, access_bookmark_on_load,
	access_fullview_mode
};
enum { resize_nn, resample }; //Resize methods
enum { option_menu, number_menu }; //Menu types

typedef struct {
	int type;
	unzFile zip_file;
	char *filename;
	char *localname;
	Uint16 current_page;
	Uint16 num_pages;
	char **page_order;
} comic_book;
typedef struct {
	char *option_text;
	int command;
	int int_data;
	char *char_data;
} event;
#ifdef PSP
typedef struct {
	Uint16 u, v;
	float x, y, z;
} vertexel16;
#endif
typedef struct {
	int type;
	SDL_Surface *screen;
	SDL_Joystick *joystick;
} device;
typedef struct {
	char zoom_box_is_active;
	int *action_status;
	SDL_Surface *page_img;
	Uint16 *offsetX;
	Uint16 *offsetY;
	float scale_ratio;
	Uint8 rotation;
	SDL_Surface *scaled_page;
} zoom_box_info;
typedef struct {
	SDL_Surface *new_page;
	Uint16 page;
	char loaded;
	comic_book *book;
} precache_info;

#include "xmlfuncs.hpp" //XML functions must be included after typedefs

inline void sdl_error();

void init_video(Uint16 w, Uint16 h);
void reset_gu();
device access_device(int device_code);
int* access_int_global(int global_code);
int set_clock(int new_clock);
int init_zoom_box(void *zbi);
void flip_screen(char flip_text);
void clear_screen();

char show_image(SDL_Surface *surf, Uint16 offsetX, Uint16 offsetY, char flip);
char show_image2(SDL_Surface *surf, SDL_Rect *src, SDL_Rect *dst, char rotation, char flip);
SDL_Surface* load_background();
char show_popup(int which);

//Make sure surface is locked first! Functions do not do boundary checking!!
inline void* access_pixel(const SDL_Surface* const surf, const Uint16 x, const Uint16 y);
inline void copy_pixel(void const *src, void *dst, Uint8 bpp);
//This function is little endian ONLY currently
inline void blend(void *pixel, SDL_PixelFormat *fmt, Uint8 r, Uint8 g, Uint8 b, Uint8 a);

#ifndef USE_GFX
SDL_Surface* resize(SDL_Surface *old, SDL_Rect *clip, Uint16 newW, Uint16 newH);
SDL_Surface* rotate(SDL_Surface *old, const char amount);
#endif
SDL_Surface* rotozoom(SDL_Surface *old, SDL_Rect *clip, Uint16 newW, Uint16 newH, const char rotation);
int copy_palette(SDL_Surface *old, SDL_Surface *new_surf);
void blend_rect(SDL_Surface *surf, SDL_Rect rect, Uint8 r, Uint8 g, Uint8 b, Uint8 a);
void draw_zoom_box(SDL_Surface *page_img, float scale_ratio, Uint8 rotation, Sint32 x, Sint32 y, Uint16 offsetX, Uint16 offsetY, Uint8 zoom_level);

//Make sure to free memory afterward!
inline char* init_block(Uint16 kb);
inline event init_event(int ev);
inline char ends_with(const char *reference, const char *against);
inline void setXY(int x, int y);
inline void next_line();
inline char* make_string(const char *original);
inline void say_left(const char *string);
inline char* ellipsis(char *string, Uint16 offset);

comic_book open_comic_book(char *dir, char *filename);
void close_comic_book(comic_book book);
comic_book goto_next_page(comic_book book);
comic_book goto_prev_page(comic_book book);
comic_book goto_page(comic_book book, Uint16 page);
event show_page(SDL_Surface *page_img, comic_book *book, Uint8 *rotation, float *scale_ratio);
event show_book(char *dir, char *book_name);
int cache_page(void *pci);

//Please note! Some of the actions are handled inside of the menu, and their
//return indicates merely that this option was selected. Do not try to perorm
//the action indicated again. These actions are: change_resize_method and
//clock_adjust
event show_menu(comic_book *book);
inline void free_event(event cmd);
Uint16 file_list(char *dir, event **ptr_to_list, char get_dirs);
char* next_file_name(char *dir, char *cur_file);
char* prev_file_name(char *dir, char *cur_file);
char dirent_is_dir(struct dirent *filent, char *dir);
void battery_meter();
int battery_meter_thread(void *is_running);

char* extract_file(comic_book *book, Uint16 page, size_t *size);
char* extract_file_from_zip(unzFile zip, size_t *size);

int handle_input(char wait);

inline int cb_compare(const void* foo, const void* bar);
inline int event_compare(const void* foo, const void* bar);

#ifdef __cplusplus
}
#endif

#endif
